import os

import matplotlib
matplotlib.use('Agg')

import numpy as np

from astropy.table import Table
from astropy.tests.helper import pytest
from astropy.io import fits
from .helpers import generate_wcs
from .. import FITSFigure

HEADER = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/2d_fits', '1904-66_TAN.hdr')

tab = Table({'RA':[347.,349.], 'DEC':[-68.,-68]})

GOOD_INPUT = [ [1,2],
               [[1,2],[3,4]],
               [np.arange(2), np.arange(2)],
               [tab['RA'], tab['DEC']]
             ]

BAD_INPUT = [ [1,['s', 'e']],
              [np.arange(2), np.sum],
              [tab['RA'], 'ewr']
            ]

@pytest.mark.parametrize(('inputval'), GOOD_INPUT)
def test_pixel_coords(inputval):
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.show_markers(inputval[0], inputval[1])
    f.close()

@pytest.mark.parametrize(('inputval'), GOOD_INPUT)
def test_wcs_coords(inputval):
    wcs = generate_wcs(HEADER)
    header = fits.Header.fromtextfile(HEADER)
    wcs.naxis1 = header['NAXIS1']
    wcs.naxis2 = header['NAXIS2']
    f = FITSFigure(wcs)
    f.show_markers(inputval[0], inputval[1])
    f.close()

@pytest.mark.parametrize(('inputval'), BAD_INPUT)
def test_pixel_coords_bad(inputval):
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    with pytest.raises(Exception) as exc:
        f.show_markers(inputval[0], inputval[1])
    assert exc.value.args[0] == "world2pix should be provided either with two scalars, two lists, or two numpy arrays"
    f.close()

@pytest.mark.parametrize(('inputval'), BAD_INPUT)
def test_wcs_coords_bad(inputval):
    wcs = generate_wcs(HEADER)
    header = fits.Header.fromtextfile(HEADER)
    wcs.naxis1 = header['NAXIS1']
    wcs.naxis2 = header['NAXIS2']
    f = FITSFigure(wcs)
    with pytest.raises(Exception) as exc:
        f.show_markers(inputval[0], inputval[1])
    f.close()
    assert exc.value.args[0] == "world2pix should be provided either with two scalars, two lists, or two numpy arrays"





from __future__ import absolute_import, print_function, division

from .decorators import auto_refresh


class Frame(object):

    @auto_refresh
    def __init__(self, parent):

        self._ax1 = parent._ax1
        self._ax2 = parent._ax2
        self._figure = parent._figure

        # Save plotting parameters (required for @auto_refresh)
        self._parameters = parent._parameters

    @auto_refresh
    def set_linewidth(self, linewidth):
        '''
        Set line width of the frame.

        Parameters
        ----------
        linewidth:
            The linewidth to use for the frame.
        '''
        for key in self._ax1.spines:
            self._ax1.spines[key].set_linewidth(linewidth)
        for key in self._ax2.spines:
            self._ax2.spines[key].set_linewidth(linewidth)

    @auto_refresh
    def set_color(self, color):
        '''
        Set color of the frame.

        Parameters
        ----------
        color:
            The color to use for the frame.
        '''
        for key in self._ax1.spines:
            self._ax1.spines[key].set_edgecolor(color)
        for key in self._ax2.spines:
            self._ax2.spines[key].set_edgecolor(color)



import os

import matplotlib
matplotlib.use('Agg')

import numpy as np
from astropy.tests.helper import pytest
from astropy.io import fits
from astropy.wcs import WCS as AstropyWCS
import astropy.utils.exceptions as aue

from .helpers import generate_file, generate_hdu, generate_wcs
from .. import FITSFigure

# The tests in this file check that the initialization and basic plotting do
# not crash for FITS files with 2 dimensions. No reference images are
# required here.

header_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/2d_fits')

HEADERS = [os.path.join(header_dir, '1904-66_AIR.hdr'),
           os.path.join(header_dir, '1904-66_AIT.hdr'),
           os.path.join(header_dir, '1904-66_ARC.hdr'),
           os.path.join(header_dir, '1904-66_AZP.hdr'),
           os.path.join(header_dir, '1904-66_BON.hdr'),
           os.path.join(header_dir, '1904-66_CAR.hdr'),
           os.path.join(header_dir, '1904-66_CEA.hdr'),
           os.path.join(header_dir, '1904-66_COD.hdr'),
           os.path.join(header_dir, '1904-66_COE.hdr'),
           os.path.join(header_dir, '1904-66_COO.hdr'),
           os.path.join(header_dir, '1904-66_COP.hdr'),
           os.path.join(header_dir, '1904-66_CSC.hdr'),
           os.path.join(header_dir, '1904-66_CYP.hdr'),
           os.path.join(header_dir, '1904-66_HPX.hdr'),
           os.path.join(header_dir, '1904-66_MER.hdr'),
           os.path.join(header_dir, '1904-66_MOL.hdr'),
           os.path.join(header_dir, '1904-66_NCP.hdr'),
           os.path.join(header_dir, '1904-66_PAR.hdr'),
           os.path.join(header_dir, '1904-66_PCO.hdr'),
           os.path.join(header_dir, '1904-66_QSC.hdr'),
           os.path.join(header_dir, '1904-66_SFL.hdr'),
           os.path.join(header_dir, '1904-66_SIN.hdr'),
           os.path.join(header_dir, '1904-66_STG.hdr'),
           os.path.join(header_dir, '1904-66_SZP.hdr'),
           os.path.join(header_dir, '1904-66_TAN.hdr'),
           os.path.join(header_dir, '1904-66_TSC.hdr'),
           os.path.join(header_dir, '1904-66_ZEA.hdr'),
           os.path.join(header_dir, '1904-66_ZPN.hdr')]

REFERENCE = os.path.join(header_dir, '1904-66_TAN.hdr')

CAR_REFERENCE = os.path.join(header_dir, '1904-66_CAR.hdr')

VALID_DIMENSIONS = [(0, 1), (1, 0)]
INVALID_DIMENSIONS = [None, (1,), (0, 2), (-4, 2), (1, 1), (2, 2), (1, 2, 3)]


# Test initialization through a filename
def test_file_init(tmpdir):
    filename = generate_file(REFERENCE, str(tmpdir))
    f = FITSFigure(filename)
    f.show_grayscale()
    f.close()


# Test initialization through an HDU object
def test_hdu_init():
    hdu = generate_hdu(REFERENCE)
    f = FITSFigure(hdu)
    f.show_grayscale()
    f.close()


# Test initialization through a WCS object
def test_wcs_init():
    wcs = generate_wcs(REFERENCE)
    if hasattr(wcs,'naxis1'):
        f = FITSFigure(wcs)
        f.show_grayscale()
        f.close()
    else:
        with pytest.raises(aue.AstropyDeprecationWarning):
            f = FITSFigure(wcs)


# Test initialization through a WCS object with wcs.to_header() as a go-between
# specifically for testing the cd -> pc -> cd hack, and has particular importance
# for AVM-generated headers
def test_wcs_toheader_init():
    wcs = generate_wcs(REFERENCE)
    header_ = fits.Header.fromtextfile(REFERENCE)
    header = wcs.to_header()
    wcs2 = AstropyWCS(header)
    wcs2.naxis1 = wcs.naxis1 = header_['NAXIS1']
    wcs2.naxis2 = wcs.naxis2 = header_['NAXIS2']
    f = FITSFigure(wcs2)
    f.show_grayscale()
    f.add_grid()
    f.close()


# Test initialization through an HDU object (no WCS)
def test_hdu_nowcs_init():
    data = np.zeros((16, 16))
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale()
    f.close()


# Test initalization through a Numpy array (no WCS)
def test_numpy_nowcs_init():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.show_grayscale()
    f.close()


# Now check initialization with valid and invalid dimensions. We just need to
# tes with HDU objects since we already tested that reading from files is ok.

# Test initialization with valid dimensions
@pytest.mark.parametrize(('dimensions'), VALID_DIMENSIONS)
def test_init_dimensions_valid(dimensions):
    hdu = generate_hdu(REFERENCE)
    f = FITSFigure(hdu, dimensions=dimensions)
    f.show_grayscale()
    f.close()

# Test initialization with invalid dimensions


@pytest.mark.parametrize(('dimensions'), INVALID_DIMENSIONS)
def test_init_dimensions_invalid(dimensions):
    hdu = generate_hdu(REFERENCE)
    with pytest.raises(ValueError):
        FITSFigure(hdu, dimensions=dimensions)

# Now check initialization of different WCS projections, and we check only
# valid dimensions

valid_parameters = []
for h in HEADERS:
    for d in VALID_DIMENSIONS:
        valid_parameters.append((h, d))


@pytest.mark.parametrize(('header', 'dimensions'), valid_parameters)
def test_init_extensive_wcs(header, dimensions):
    hdu = generate_hdu(header)
    if 'CAR' in header:
        f = FITSFigure(hdu, dimensions=dimensions, convention='calabretta')
    else:
        f = FITSFigure(hdu, dimensions=dimensions)
    f.show_grayscale()
    f.add_grid()
    f.close()

# Check that for CAR projections, an exception is raised if no convention is specified


@pytest.mark.parametrize(('dimensions'), VALID_DIMENSIONS)
def test_init_car_invalid(dimensions):
    hdu = generate_hdu(CAR_REFERENCE)
    with pytest.raises(Exception):
        FITSFigure(hdu, dimensions=dimensions)

# Check that images containing only NaN or Inf values don't crash FITSFigure
def test_init_only_naninf():
    data = np.ones((10,10)) * np.nan
    data[::2,::2] = np.inf
    f = FITSFigure(data)
    f.show_grayscale()
    f.show_colorscale()


def test_init_single_pixel():
    data = np.zeros((4,4))
    data[...] = np.nan
    data[2,2] = 1
    f = FITSFigure(data)
    f.show_grayscale()



import os
import shutil
import tempfile

import numpy as np
from matplotlib.testing.compare import compare_images
from astropy.tests.helper import pytest
from .. import FITSFigure
from .helpers import generate_file


class BaseImageTests(object):

    @classmethod
    def setup_class(cls):

        cls._moduledir = os.path.dirname(__file__)
        cls._data_dir = os.path.abspath(os.path.join(cls._moduledir, 'data'))
        cls._baseline_images_dir = os.path.abspath(os.path.join(cls._moduledir, 'baseline_images'))

        header_1 = os.path.join(cls._data_dir, '2d_fits/1904-66_AIR.hdr')
        cls.filename_1 = generate_file(header_1, str(tempfile.mkdtemp()))

        header_2 = os.path.join(cls._data_dir, '2d_fits/2MASS_k.hdr')
        cls.filename_2 = generate_file(header_2, str(tempfile.mkdtemp()))

        header_3 = os.path.join(cls._data_dir, '3d_fits/cube.hdr')
        cls.filename_3 = generate_file(header_3, str(tempfile.mkdtemp()))

    # method to create baseline or test images
    def generate_or_test(self, generate, figure, image, adjust_bbox=True, tolerance=1):
        if generate is None:
            result_dir = tempfile.mkdtemp()
            test_image = os.path.abspath(os.path.join(result_dir, image))

            # distutils will put the baseline images in non-accessible places,
            # copy to our tmpdir to be sure to keep them in case of failure
            orig_baseline_image = os.path.abspath(os.path.join(self._baseline_images_dir, image))
            baseline_image = os.path.abspath(os.path.join(result_dir, 'baseline-'+image))
            shutil.copyfile(orig_baseline_image, baseline_image)

            figure.save(test_image)

            if not os.path.exists(baseline_image):
                raise Exception("""Image file not found for comparision test
                                Generated Image:
                                \t{test}
                                This is expected for new tests.""".format(
                                    test=test_image))

            msg = compare_images(baseline_image, test_image, tol=tolerance)

            if msg is None:
                shutil.rmtree(result_dir)
            else:
                raise Exception(msg)
        else:
            figure.save(os.path.abspath(os.path.join(generate, image)), adjust_bbox=adjust_bbox)
            pytest.skip("Skipping test, since generating data")


class TestBasic(BaseImageTests):

    # Test for showing grayscale
    def test_basic_image(self, generate):
        f = FITSFigure(self.filename_2)
        f.show_grayscale()
        self.generate_or_test(generate, f, 'basic_image.png')
        f.close()

    def test_ticks_labels_options(self, generate):
        f = FITSFigure(self.filename_2)
        f.ticks.set_color('black')
        f.axis_labels.set_xposition('top')
        f.axis_labels.set_yposition('right')
        f.axis_labels.set_font(size='medium', weight='medium',
                               stretch='normal', style='normal')
        f.tick_labels.set_xformat('dd:mm:ss.ss')
        f.tick_labels.set_yformat('hh:mm:ss.ss')
        f.tick_labels.set_style('colons')
        f.ticks.set_xspacing(0.2)
        f.ticks.set_yspacing(0.2)
        f.ticks.set_minor_frequency(10)
        self.generate_or_test(generate, f, 'tick_labels_options.png')
        f.close()

    # Test for showing colorscale
    def test_show_colorbar_scalebar_beam(self, generate):
        f = FITSFigure(self.filename_1)
        f.ticks.set_color('black')
        f.show_colorscale(vmin=-0.1, vmax=0.1)
        f.add_colorbar()
        f.add_scalebar(7.5)
        f.add_beam(major=0.5, minor=0.2, angle=10.)
        f.tick_labels.hide()
        self.generate_or_test(generate, f, 'colorbar_scalebar_beam.png')
        f.close()

    # Test for overlaying shapes
    def test_overlay_shapes(self, generate):
        f = FITSFigure(self.filename_1)
        f.ticks.set_color('black')
        f.show_markers([360., 350., 340.], [-61., -62., -63])
        f.show_ellipses(330., -66., 0.15, 2., 10.)
        f.show_rectangles([355., 350.], [-71, -72], [0.5, 1], [2, 1])
        f.show_arrows([340., 360], [-72, -68], [2, -2], [2, 2])
        f.add_label(350., -66., 'text')
        f.add_label(0.4, 0.25, 'text', relative=True)
        f.frame.set_linewidth(1)  # points
        f.frame.set_color('black')
        f.axis_labels.hide()
        self.generate_or_test(generate, f, 'overlay_shapes.png')
        f.close()

    # Test for grid
    def test_grid(self, generate):
        f = FITSFigure(self.filename_1)
        f.ticks.set_color('black')
        f.add_grid()
        f.grid.set_color('red')
        f.grid.set_alpha(0.8)
        f.grid.set_linestyle('solid')
        f.grid.set_xspacing('tick')
        f.grid.set_yspacing(3)
        self.generate_or_test(generate, f, 'grid.png')
        f.close()

    # Test recenter
    def test_recenter(self, generate):
        f = FITSFigure(self.filename_2)
        f.ticks.set_color('black')
        f.recenter(266.5, -29.0, width=0.1, height=0.1)
        f.axis_labels.set_xpad(20)
        f.axis_labels.set_ypad(20)
        self.generate_or_test(generate, f, 'recenter.png')
        f.close()

    # Test overlaying contours
    def test_contours(self, generate):
        data = np.arange(256).reshape((16, 16))
        f = FITSFigure(data)
        f.ticks.set_color('black')
        f.show_contour(data, levels=np.linspace(1., 254., 10), filled=False)
        self.generate_or_test(generate, f, 'contours.png')
        f.close()

    # Test cube slice
    def test_cube_slice(self, generate):
        f = FITSFigure(self.filename_3, dimensions=[2, 0], slices=[10])
        f.ticks.set_color('black')
        f.add_grid()
        f.grid.set_color('black')
        f.grid.set_linestyle('solid')
        f.grid.set_xspacing(250)
        f.grid.set_yspacing(0.01)
        f.tick_labels.set_xformat('%g')
        f.tick_labels.set_yformat('dd:mm:ss.ss')
        self.generate_or_test(generate, f, 'cube_slice.png')
        f.close()




from __future__ import absolute_import, print_function, division

from astropy import log


def check(header, convention=None, dimensions=[0, 1]):

    ix = dimensions[0] + 1
    iy = dimensions[1] + 1

    # If header does not contain CTYPE keywords, assume that the WCS is
    # missing or incomplete, and replace it with a 1-to-1 pixel mapping
    if 'CTYPE%i' % ix not in header or 'CTYPE%i' % iy not in header:
        log.warning("No WCS information found in header - using pixel coordinates")
        header['CTYPE%i' % ix] = 'PIXEL'
        header['CTYPE%i' % iy] = 'PIXEL'
        header['CRVAL%i' % ix] = 0.
        header['CRVAL%i' % iy] = 0.
        header['CRPIX%i' % ix] = 0.
        header['CRPIX%i' % iy] = 0.
        header['CDELT%i' % ix] = 1.
        header['CDELT%i' % iy] = 1.

    if header['CTYPE%i' % ix][4:] == '-CAR' and header['CTYPE%i' % iy][4:] == '-CAR':

        if header['CTYPE%i' % ix][:4] == 'DEC-' or header['CTYPE%i' % ix][1:4] == 'LAT':
            ilon = iy
            ilat = ix
        elif header['CTYPE%i' % iy][:4] == 'DEC-' or header['CTYPE%i' % iy][1:4] == 'LAT':
            ilon = ix
            ilat = iy
        else:
            ilon = None
            ilat = None

        if ilat is not None and header['CRVAL%i' % ilat] != 0:

            if convention == 'calabretta':
                pass  # we don't need to do anything
            elif convention == 'wells':
                if 'CDELT%i' % ilat not in header:
                    raise Exception("Need CDELT%i to be present for wells convention" % ilat)
                crpix = header['CRPIX%i' % ilat]
                crval = header['CRVAL%i' % ilat]
                cdelt = header['CDELT%i' % ilat]
                crpix = crpix - crval / cdelt
                try:
                    header['CRPIX%i' % ilat] = crpix
                    header['CRVAL%i' % ilat] = 0.
                except:  # older versions of PyFITS
                    header.update('CRPIX%i' % ilat, crpix)
                    header.update('CRVAL%i' % ilon, 0.)

            else:
                raise Exception('''WARNING: projection is Plate Caree (-CAR) and
                CRVALy is not zero. This can be interpreted either according to
                Wells (1981) or Calabretta (2002). The former defines the
                projection as rectilinear regardless of the value of CRVALy,
                whereas the latter defines the projection as rectilinear only when
                CRVALy is zero. You will need to specify the convention to assume
                by setting either convention='wells' or convention='calabretta'
                when initializing the FITSFigure instance. ''')

    return header



from __future__ import absolute_import, print_function, division

import numpy as np
try:
    from astropy.convolution import convolve as astropy_convolve, Gaussian2DKernel, Box2DKernel
    make_kernel = None
except ImportError:
    from astropy.nddata import convolve as astropy_convolve, make_kernel


def convolve(image, smooth=3, kernel='gauss'):

    if smooth is None and kernel in ['box', 'gauss']:
        return image

    if smooth is not None and not np.isscalar(smooth):
        raise ValueError("smooth= should be an integer - for more complex "
                         "kernels, pass an array containing the kernel "
                         "to the kernel= option")

    # The Astropy convolution doesn't treat +/-Inf values correctly yet, so we
    # convert to NaN here.

    image_fixed = np.array(image, dtype=float, copy=True)
    image_fixed[np.isinf(image)] = np.nan

    if kernel == 'gauss':
        if make_kernel is None:
            kernel = Gaussian2DKernel(smooth, x_size=smooth * 5, y_size=smooth * 5)
        else:
            kernel = make_kernel((smooth * 5, smooth * 5), smooth, 'gaussian')
    elif kernel == 'box':
        if make_kernel is None:
            kernel = Box2DKernel(smooth, x_size=smooth * 5, y_size=smooth * 5)
        else:
            kernel = make_kernel((smooth * 5, smooth * 5), smooth, 'boxcar')
    else:
        kernel = kernel

    return astropy_convolve(image, kernel, boundary='extend')



import matplotlib
matplotlib.use('Agg')

import numpy as np
from astropy.tests.helper import pytest

from .. import FITSFigure

# Test simple contour generation with Numpy example


def test_numpy_downsample():
    data = np.arange(256).reshape((16, 16))
    f = FITSFigure(data, downsample=2)
    f.show_grayscale()
    f.close()



import os

import matplotlib
matplotlib.use('Agg')

import numpy as np
from astropy.tests.helper import pytest
from astropy import units as u
from astropy.io import fits

from .. import FITSFigure


header_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/2d_fits')

HEADER = fits.Header.fromtextfile(os.path.join(header_dir, '1904-66_TAN.hdr'))
HDU = fits.PrimaryHDU(np.zeros((16, 16)), HEADER)


def test_scalebar_add_invalid():
    f = FITSFigure(HDU)
    with pytest.raises(TypeError):
        f.add_scalebar()


def test_scalebar_addremove():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.remove_scalebar()
    f.add_scalebar(0.1)
    f.close()


def test_scalebar_showhide():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.hide()
    f.scalebar.show(0.1)
    f.close()


def test_scalebar_length():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.set_length(0.01)
    f.scalebar.set_length(0.1)
    f.close()


@pytest.mark.parametrize('quantity', [1*u.arcsec, u.arcsec, 2*u.degree, 5*u.radian])
def test_scalebar_length_quantity(quantity):
    f = FITSFigure(HDU)
    f.add_scalebar(quantity)
    f.scalebar.set_length(quantity)
    f.close()


def test_scalebar_label():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.set_label('1 pc')
    f.scalebar.set_label('5 AU')
    f.scalebar.set_label('2"')
    f.close()


def test_scalebar_corner():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    for corner in ['top', 'bottom', 'left', 'right', 'top left', 'top right',
                   'bottom left', 'bottom right']:
        f.scalebar.set_corner(corner)
    f.close()


def test_scalebar_frame():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.set_frame(True)
    f.scalebar.set_frame(False)
    f.close()


def test_scalebar_color():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.set_color('black')
    f.scalebar.set_color('#003344')
    f.scalebar.set_color((1.0, 0.4, 0.3))
    f.close()


def test_scalebar_alpha():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.set_alpha(0.1)
    f.scalebar.set_alpha(0.2)
    f.scalebar.set_alpha(0.5)
    f.close()


def test_scalebar_font():
    f = FITSFigure(HDU)
    f.add_scalebar(0.1)
    f.scalebar.set_font(size='small', weight='bold', stretch='normal',
                        family='serif', style='normal', variant='normal')
    f.close()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, division, unicode_literals

import warnings

import numpy as np
import matplotlib.pyplot as mpl
from matplotlib.font_manager import FontProperties

from . import wcs_util
from . import angle_util as au
from .decorators import auto_refresh, fixdocstring


class TickLabels(object):

    def __init__(self, parent):

        # Store references to axes
        self._ax1 = parent._ax1
        self._ax2 = parent._ax2
        self._wcs = parent._wcs
        self._figure = parent._figure

        # Save plotting parameters (required for @auto_refresh)
        self._parameters = parent._parameters

        # Set font
        self._label_fontproperties = FontProperties()

        self.set_style('plain')

        system, equinox, units = wcs_util.system(self._wcs)

        # Set default label format
        if self._wcs.xaxis_coord_type in ['longitude', 'latitude']:
            if system['name'] == 'equatorial':
                if self._wcs.xaxis_coord_type == 'longitude':
                    self.set_xformat("hh:mm:ss.ss")
                else:
                    self.set_xformat("dd:mm:ss.s")
            else:
                self.set_xformat("ddd.dddd")
        else:
            self.set_xformat('%g')

        if self._wcs.yaxis_coord_type in ['longitude', 'latitude']:
            if system['name'] == 'equatorial':
                if self._wcs.yaxis_coord_type == 'longitude':
                    self.set_yformat("hh:mm:ss.ss")
                else:
                    self.set_yformat("dd:mm:ss.s")
            else:
                self.set_yformat("ddd.dddd")
        else:
            self.set_yformat('%g')

        # Set major tick formatters
        fx1 = WCSFormatter(wcs=self._wcs, coord='x')
        fy1 = WCSFormatter(wcs=self._wcs, coord='y')
        self._ax1.xaxis.set_major_formatter(fx1)
        self._ax1.yaxis.set_major_formatter(fy1)

        fx2 = mpl.NullFormatter()
        fy2 = mpl.NullFormatter()
        self._ax2.xaxis.set_major_formatter(fx2)
        self._ax2.yaxis.set_major_formatter(fy2)

        # Cursor display
        self._ax1._cursor_world = True
        self._figure.canvas.mpl_connect('key_press_event', self._set_cursor_prefs)

    @auto_refresh
    def set_xformat(self, format):
        '''
        Set the format of the x-axis tick labels.

        If the x-axis type is ``longitude`` or ``latitude``, then the options
        are:

            * ``ddd.ddddd`` - decimal degrees, where the number of decimal places can be varied
            * ``hh`` or ``dd`` - hours (or degrees)
            * ``hh:mm`` or ``dd:mm`` - hours and minutes (or degrees and arcminutes)
            * ``hh:mm:ss`` or ``dd:mm:ss`` - hours, minutes, and seconds (or degrees, arcminutes, and arcseconds)
            * ``hh:mm:ss.ss`` or ``dd:mm:ss.ss`` - hours, minutes, and seconds (or degrees, arcminutes, and arcseconds), where the number of decimal places can be varied.

        If the x-axis type is ``scalar``, then the format should be a valid
        python string format beginning with a ``%``.

        If one of these arguments is not specified, the format for that axis
        is left unchanged.
        '''
        if self._wcs.xaxis_coord_type in ['longitude', 'latitude']:
            if format.startswith('%'):
                raise Exception("Cannot specify Python format for longitude or latitude")
            try:
                if not self._ax1.xaxis.apl_auto_tick_spacing:
                    au._check_format_spacing_consistency(format, self._ax1.xaxis.apl_tick_spacing)
            except au.InconsistentSpacing:
                warnings.warn("WARNING: Requested label format is not accurate enough to display ticks. The label format will not be changed.")
                return
        else:
            if not format.startswith('%'):
                raise Exception("For scalar tick labels, format should be a Python format beginning with %")

        self._ax1.xaxis.apl_label_form = format
        self._ax2.xaxis.apl_label_form = format

    @auto_refresh
    def set_yformat(self, format):
        '''
        Set the format of the y-axis tick labels.

        If the y-axis type is ``longitude`` or ``latitude``, then the options
        are:

            * ``ddd.ddddd`` - decimal degrees, where the number of decimal places can be varied
            * ``hh`` or ``dd`` - hours (or degrees)
            * ``hh:mm`` or ``dd:mm`` - hours and minutes (or degrees and arcminutes)
            * ``hh:mm:ss`` or ``dd:mm:ss`` - hours, minutes, and seconds (or degrees, arcminutes, and arcseconds)
            * ``hh:mm:ss.ss`` or ``dd:mm:ss.ss`` - hours, minutes, and seconds (or degrees, arcminutes, and arcseconds), where the number of decimal places can be varied.

        If the y-axis type is ``scalar``, then the format should be a valid
        python string format beginning with a ``%``.

        If one of these arguments is not specified, the format for that axis
        is left unchanged.
        '''
        if self._wcs.yaxis_coord_type in ['longitude', 'latitude']:
            if format.startswith('%'):
                raise Exception("Cannot specify Python format for longitude or latitude")
            try:
                if not self._ax1.yaxis.apl_auto_tick_spacing:
                    au._check_format_spacing_consistency(format, self._ax1.yaxis.apl_tick_spacing)
            except au.InconsistentSpacing:
                warnings.warn("WARNING: Requested label format is not accurate enough to display ticks. The label format will not be changed.")
                return
        else:
            if not format.startswith('%'):
                raise Exception("For scalar tick labels, format should be a Python format beginning with %")

        self._ax1.yaxis.apl_label_form = format
        self._ax2.yaxis.apl_label_form = format

    @auto_refresh
    def set_style(self, style):
        """
        Set the format of the x-axis tick labels.

        This can be 'colons' or 'plain':

            * 'colons' uses colons as separators, for example 31:41:59.26 +27:18:28.1
            * 'plain' uses letters and symbols as separators, for example 31h41m59.26s +27º18'28.1"
        """

        if style == 'latex':
            warnings.warn("latex has now been merged with plain - whether or not to use LaTeX is controlled through set_system_latex")
            style = 'plain'

        if not style in ['colons', 'plain']:
            raise Exception("Label style should be one of colons/plain")

        self._ax1.xaxis.apl_labels_style = style
        self._ax1.yaxis.apl_labels_style = style
        self._ax2.xaxis.apl_labels_style = style
        self._ax2.yaxis.apl_labels_style = style

    @auto_refresh
    @fixdocstring
    def set_font(self, family=None, style=None, variant=None, stretch=None, weight=None, size=None, fontproperties=None):
        """
        Set the font of the tick labels.

        Parameters
        ----------

        common: family, style, variant, stretch, weight, size, fontproperties

        Notes
        -----

        Default values are set by matplotlib or previously set values if
        set_font has already been called. Global default values can be set by
        editing the matplotlibrc file.
        """

        if family:
            self._label_fontproperties.set_family(family)

        if style:
            self._label_fontproperties.set_style(style)

        if variant:
            self._label_fontproperties.set_variant(variant)

        if stretch:
            self._label_fontproperties.set_stretch(stretch)

        if weight:
            self._label_fontproperties.set_weight(weight)

        if size:
            self._label_fontproperties.set_size(size)

        if fontproperties:
            self._label_fontproperties = fontproperties

        for tick in self._ax1.get_xticklabels():
            tick.set_fontproperties(self._label_fontproperties)
        for tick in self._ax1.get_yticklabels():
            tick.set_fontproperties(self._label_fontproperties)
        for tick in self._ax2.get_xticklabels():
            tick.set_fontproperties(self._label_fontproperties)
        for tick in self._ax2.get_yticklabels():
            tick.set_fontproperties(self._label_fontproperties)

    @auto_refresh
    def show(self):
        """
        Show the x- and y-axis tick labels.
        """
        self.show_x()
        self.show_y()

    @auto_refresh
    def hide(self):
        """
        Hide the x- and y-axis tick labels.
        """
        self.hide_x()
        self.hide_y()

    @auto_refresh
    def show_x(self):
        """
        Show the x-axis tick labels.
        """

        for tick in self._ax1.get_xticklabels():
            tick.set_visible(True)
        for tick in self._ax2.get_xticklabels():
            tick.set_visible(True)

    @auto_refresh
    def hide_x(self):
        """
        Hide the x-axis tick labels.
        """

        for tick in self._ax1.get_xticklabels():
            tick.set_visible(False)
        for tick in self._ax2.get_xticklabels():
            tick.set_visible(False)

    @auto_refresh
    def show_y(self):
        """
        Show the y-axis tick labels.
        """

        for tick in self._ax1.get_yticklabels():
            tick.set_visible(True)
        for tick in self._ax2.get_yticklabels():
            tick.set_visible(True)

    @auto_refresh
    def hide_y(self):
        """
        Hide the y-axis tick labels.
        """

        for tick in self._ax1.get_yticklabels():
            tick.set_visible(False)
        for tick in self._ax2.get_yticklabels():
            tick.set_visible(False)

    @auto_refresh
    def set_xposition(self, position):
        """
        Set the position of the x-axis tick labels ('top' or 'bottom')
        """
        if position == 'bottom':
            fx1 = WCSFormatter(wcs=self._wcs, coord='x')
            self._ax1.xaxis.set_major_formatter(fx1)
            fx2 = mpl.NullFormatter()
            self._ax2.xaxis.set_major_formatter(fx2)
        elif position == 'top':
            fx1 = mpl.NullFormatter()
            self._ax1.xaxis.set_major_formatter(fx1)
            fx2 = WCSFormatter(wcs=self._wcs, coord='x')
            self._ax2.xaxis.set_major_formatter(fx2)
        else:
            raise ValueError("position should be one of 'top' or 'bottom'")

    @auto_refresh
    def set_yposition(self, position):
        """
        Set the position of the y-axis tick labels ('left' or 'right')
        """
        if position == 'left':
            fy1 = WCSFormatter(wcs=self._wcs, coord='y')
            self._ax1.yaxis.set_major_formatter(fy1)
            fy2 = mpl.NullFormatter()
            self._ax2.yaxis.set_major_formatter(fy2)
        elif position == 'right':
            fy1 = mpl.NullFormatter()
            self._ax1.yaxis.set_major_formatter(fy1)
            fy2 = WCSFormatter(wcs=self._wcs, coord='y')
            self._ax2.yaxis.set_major_formatter(fy2)
        else:
            raise ValueError("position should be one of 'left' or 'right'")

    def _set_cursor_prefs(self, event, **kwargs):
        if event.key == 'c':
            self._ax1._cursor_world = not self._ax1._cursor_world

    def _cursor_position(self, x, y):

        xaxis = self._ax1.xaxis
        yaxis = self._ax1.yaxis

        if self._ax1._cursor_world:

            xw, yw = wcs_util.pix2world(self._wcs, x, y)

            if self._wcs.xaxis_coord_type in ['longitude', 'latitude']:

                xw = au.Angle(degrees=xw, latitude=self._wcs.xaxis_coord_type == 'latitude')

                hours = 'h' in xaxis.apl_label_form

                if hours:
                    xw = xw.tohours()

                if xaxis.apl_labels_style in ['plain', 'latex']:
                    sep = ('d', 'm', 's')
                    if hours:
                        sep = ('h', 'm', 's')
                elif xaxis.apl_labels_style == 'colons':
                    sep = (':', ':', '')

                xlabel = xw.tostringlist(format=xaxis.apl_label_form, sep=sep)
                xlabel = "".join(xlabel)

            else:

                xlabel = xaxis.apl_label_form % xw

            if self._wcs.yaxis_coord_type in ['longitude', 'latitude']:

                yw = au.Angle(degrees=yw, latitude=self._wcs.yaxis_coord_type == 'latitude')

                hours = 'h' in yaxis.apl_label_form

                if hours:
                    yw = yw.tohours()

                if yaxis.apl_labels_style in ['plain', 'latex']:
                    sep = ('d', 'm', 's')
                    if hours:
                        sep = ('h', 'm', 's')
                elif yaxis.apl_labels_style == 'colons':
                    sep = (':', ':', '')

                ylabel = yw.tostringlist(format=yaxis.apl_label_form, sep=sep)
                ylabel = "".join(ylabel)

            else:

                ylabel = yaxis.apl_label_form % yw

            return  "%s %s (world)" % (xlabel, ylabel)

        else:

            return "%g %g (pixel)" % (x, y)


class WCSFormatter(mpl.Formatter):

    def __init__(self, wcs=False, coord='x'):
        self._wcs = wcs
        self.coord = coord

    def __call__(self, x, pos=None):
        """
        Return the format for tick val x at position pos; pos=None indicated
        unspecified
        """
        self.coord_type = self._wcs.xaxis_coord_type if self.coord == 'x' else self._wcs.yaxis_coord_type

        if self.coord_type in ['longitude', 'latitude']:

            au._check_format_spacing_consistency(self.axis.apl_label_form, self.axis.apl_tick_spacing)

            hours = 'h' in self.axis.apl_label_form

            if self.axis.apl_labels_style == 'plain':
                if mpl.rcParams['text.usetex']:
                    label_style = 'plain_tex'
                else:
                    label_style = 'plain_notex'
            else:
                label_style = self.axis.apl_labels_style

            if label_style == 'plain_notex':
                sep = ('\u00b0', "'", '"')
                if hours:
                    sep = ('h', 'm', 's')
            elif label_style == 'colons':
                sep = (':', ':', '')
            elif label_style == 'plain_tex':
                if hours:
                    sep = ('^{h}', '^{m}', '^{s}')
                else:
                    sep = ('^{\circ}', '^{\prime}', '^{\prime\prime}')

            ipos = np.argmin(np.abs(self.axis.apl_tick_positions_pix - x))

            label = self.axis.apl_tick_spacing * self.axis.apl_tick_positions_world[ipos]
            if hours:
                label = label.tohours()
            label = label.tostringlist(format=self.axis.apl_label_form, sep=sep)

            # Check if neighboring label is similar and if so whether some
            # elements of the current label are redundant and can be dropped.
            # This should only be done for sexagesimal coordinates

            if len(label) > 1:

                if self.coord == x or self.axis.apl_tick_positions_world[ipos] > 0:
                    comp_ipos = ipos - 1
                else:
                    comp_ipos = ipos + 1

                if comp_ipos >= 0 and comp_ipos <= len(self.axis.apl_tick_positions_pix) - 1:

                    comp_label = self.axis.apl_tick_spacing * self.axis.apl_tick_positions_world[comp_ipos]
                    if hours:
                        comp_label = comp_label.tohours()
                    comp_label = comp_label.tostringlist(format=self.axis.apl_label_form, sep=sep)

                    for iter in range(len(label)):
                        if comp_label[0] == label[0]:
                            label.pop(0)
                            comp_label.pop(0)
                        else:
                            break

        else:

            ipos = np.argmin(np.abs(self.axis.apl_tick_positions_pix - x))
            label = self.axis.apl_tick_spacing * self.axis.apl_tick_positions_world[ipos]
            label = self.axis.apl_label_form % label

        if mpl.rcParams['text.usetex']:
            return "$" + "".join(label) + "$"
        else:
            return "".join(label)



from __future__ import absolute_import, print_function, division

import numpy as np

from . import math_util


def smart_round_angle_decimal(x, latitude=False):

    x = np.log10(x)
    e = np.floor(x)
    x -= e
    x = 10. ** x
    divisors_10 = math_util.divisors(10)
    x = math_util.closest(divisors_10, x)
    x = x * 10. ** e

    return x


def _get_label_precision(format):

    if format[0] == "%":
        if "i" in format:
            return 1
        elif "f" in format:
            return 10. ** (-len((format % 1).split('.')[1]))
        elif "e" in format:
            return None  # need to figure this out
        else:
            return None  # need to figure this out
    elif "." in format:
        return 10 ** (-len(format.split('.')[1]))
    else:
        return 1


class InconsistentSpacing(Exception):
    pass


def _check_format_spacing_consistency(format, spacing):
    '''
    Check whether the format can correctly show labels with the specified
    spacing.

    For example, if the tick spacing is set to 1 arcsecond, but the format is
    set to dd:mm, then the labels cannot be correctly shown. Similarly, if the
    spacing is set to 1/1000 of a degree, or 3.6", then a format of dd:mm:ss
    will cause rounding errors, because the spacing includes fractional
    arcseconds.

    This function will raise a warning if the format and spacing are
    inconsistent.
    '''

    label_spacing = _get_label_precision(format)

    if label_spacing is None:
        return  # can't determine minimum spacing, so can't check

    if spacing % label_spacing != 0.:
        raise InconsistentSpacing('Label format and tick spacing are inconsistent. Make sure that the tick spacing is a multiple of the smallest angle that can be represented by the specified format (currently %s). For example, if the format is dd:mm:ss.s, then the tick spacing has to be a multiple of 0.1". Similarly, if the format is hh:mm:ss, then the tick spacing has to be a multiple of 15". If you got this error as a result of interactively zooming in to a small region, this means that the default display format for the labels is not accurate enough, so you will need to increase the format precision.' % format)



# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2015-02-15 18:03:36.494515
from __future__ import unicode_literals
import datetime

version = "0.4.5"
githash = "None"


major = 0
minor = 4
bugfix = 5

release = True
timestamp = datetime.datetime(2015, 2, 15, 18, 3, 36, 494515)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



from __future__ import absolute_import, print_function, division

# The APLpyNormalize class is largely based on code provided by Sarah Graves.

import numpy as np
import numpy.ma as ma

import matplotlib.cbook as cbook
from matplotlib.colors import Normalize


class APLpyNormalize(Normalize):
    '''
    A Normalize class for imshow that allows different stretching functions
    for astronomical images.
    '''

    def __init__(self, stretch='linear', exponent=5, vmid=None, vmin=None,
                 vmax=None, clip=False):
        '''
        Initalize an APLpyNormalize instance.

        Parameters
        ----------

        vmin : None or float, optional
            Minimum pixel value to use for the scaling.

        vmax : None or float, optional
            Maximum pixel value to use for the scaling.

        stretch : { 'linear', 'log', 'sqrt', 'arcsinh', 'power' }, optional
            The stretch function to use (default is 'linear').

        vmid : None or float, optional
            Mid-pixel value used for the log and arcsinh stretches. If
            set to None, a default value is picked.

        exponent : float, optional
            if self.stretch is set to 'power', this is the exponent to use.

        clip : str, optional
            If clip is True and the given value falls outside the range,
            the returned value will be 0 or 1, whichever is closer.
        '''

        if vmax < vmin:
            raise Exception("vmax should be larger than vmin")

        # Call original initalization routine
        Normalize.__init__(self, vmin=vmin, vmax=vmax, clip=clip)

        # Save parameters
        self.stretch = stretch
        self.exponent = exponent

        if stretch == 'power' and np.equal(self.exponent, None):
            raise Exception("For stretch=='power', an exponent should be specified")

        if np.equal(vmid, None):
            if stretch == 'log':
                if vmin > 0:
                    self.midpoint = vmax / vmin
                else:
                    raise Exception("When using a log stretch, if vmin < 0, then vmid has to be specified")
            elif stretch == 'arcsinh':
                self.midpoint = -1. / 30.
            else:
                self.midpoint = None
        else:
            if stretch == 'log':
                if vmin < vmid:
                    raise Exception("When using a log stretch, vmin should be larger than vmid")
                self.midpoint = (vmax - vmid) / (vmin - vmid)
            elif stretch == 'arcsinh':
                self.midpoint = (vmid - vmin) / (vmax - vmin)
            else:
                self.midpoint = None

    def __call__(self, value, clip=None):

        #read in parameters
        method = self.stretch
        exponent = self.exponent
        midpoint = self.midpoint

        # ORIGINAL MATPLOTLIB CODE

        if clip is None:
            clip = self.clip

        if cbook.iterable(value):
            vtype = 'array'
            val = ma.asarray(value).astype(np.float)
        else:
            vtype = 'scalar'
            val = ma.array([value]).astype(np.float)

        self.autoscale_None(val)
        vmin, vmax = self.vmin, self.vmax
        if vmin > vmax:
            raise ValueError("minvalue must be less than or equal to maxvalue")
        elif vmin == vmax:
            return 0.0 * val
        else:
            if clip:
                mask = ma.getmask(val)
                val = ma.array(np.clip(val.filled(vmax), vmin, vmax),
                                mask=mask)
            result = (val - vmin) * (1.0 / (vmax - vmin))

            # CUSTOM APLPY CODE

            # Keep track of negative values
            negative = result < 0.

            if self.stretch == 'linear':

                pass

            elif self.stretch == 'log':

                result = ma.log10(result * (self.midpoint - 1.) + 1.) \
                       / ma.log10(self.midpoint)

            elif self.stretch == 'sqrt':

                result = ma.sqrt(result)

            elif self.stretch == 'arcsinh':

                result = ma.arcsinh(result / self.midpoint) \
                       / ma.arcsinh(1. / self.midpoint)

            elif self.stretch == 'power':

                result = ma.power(result, exponent)

            else:

                raise Exception("Unknown stretch in APLpyNormalize: %s" %
                                self.stretch)

            # Now set previously negative values to 0, as these are
            # different from true NaN values in the FITS image
            result[negative] = -np.inf

        if vtype == 'scalar':
            result = result[0]

        return result

    def inverse(self, value):

        # ORIGINAL MATPLOTLIB CODE

        if not self.scaled():
            raise ValueError("Not invertible until scaled")

        vmin, vmax = self.vmin, self.vmax

        # CUSTOM APLPY CODE

        if cbook.iterable(value):
            val = ma.asarray(value)
        else:
            val = value

        if self.stretch == 'linear':

            pass

        elif self.stretch == 'log':

            val = (ma.power(10., val * ma.log10(self.midpoint)) - 1.) / (self.midpoint - 1.)

        elif self.stretch == 'sqrt':

            val = val * val

        elif self.stretch == 'arcsinh':

            val = self.midpoint * \
                  ma.sinh(val * ma.arcsinh(1. / self.midpoint))

        elif self.stretch == 'power':

            val = ma.power(val, (1. / self.exponent))

        else:

            raise Exception("Unknown stretch in APLpyNormalize: %s" %
                            self.stretch)

        return vmin + val * (vmax - vmin)



from __future__ import absolute_import, print_function, division

import math
import struct

import numpy as np
from . import math_util


def almost_equal(a, b):
    c = struct.pack("<dd", a, b)
    d = struct.unpack("<qq", c)
    diff = abs(d[1] - d[0])
    return diff < 100


class Angle(object):

    def __init__(self, degrees='none', sexagesimal='none', latitude=False):

        if degrees != 'none':

            # Find out if the angle is negative
            negative = degrees < 0

            # Treat angle as positive
            degrees = np.abs(degrees)

            # Decompose angle into degrees, minutes, seconds
            m, d = math.modf(degrees)
            s, m = math.modf(m * 60.)
            s = s * 60.

            # Express degrees and minutes as integers
            d, m = int(round(d)), int(round(m))

            # Put back minus sign if negative
            if negative:
                d, m, s = -d, -m, -s

            # Set angle to tuple of degrees/minutes/seconds
            self.angle = (d, m, s)

        elif sexagesimal != 'none':
            self.angle = sexagesimal

        # Whether to keep the angle between 0 and 360 or -90 and 90
        self.latitude = latitude

        self.negative = False

        self._simplify()

    def _simplify(self):

        # Decompose angle
        d, m, s = self.angle

        # Make sure seconds are between 0. (inclusive) and 60. (exclusive)
        r = np.mod(s, 60.)
        m = m + int(round(s - r)) / 60
        s = r

        # Make sure minutes are between 0 and 59 (both inclusive)
        r = np.mod(m, 60)
        d = d + (m - r) / 60
        m = r

        # Make sure degrees are between 0 and 359 (both inclusive)
        d = np.mod(d, 360)

        # If angle is latitude, then:
        # - if degrees are between 90 and 270, angle is invalid
        # - if angle is between 270 and 360, subtract 360 degrees

        if self.latitude and d > 90:
            if d >= 270:
                self.negative = True
                d, m, s = 359 - d, 59 - m, 60. - s
                if s == 60.:
                    s = s - 60.
                    m = m + 1
                if m == 60:
                    m = m - 60.
                    d = d + 1
            else:
                raise Exception("latitude should be between -90 and 90 \
                    degrees")

        # Set new angle
        self.angle = (d, m, s)

    def todegrees(self):
        d, m, s = self.angle
        degrees = d + m / 60. + s / 3600.
        if self.negative:
            degrees = - degrees
        return degrees

    def tohours(self):

        d, m, s = self.angle

        rd = np.mod(d, 15)
        h = (d - rd) / 15

        rm = np.mod(m, 15)
        m = (m - rm) / 15 + rd * 4

        s = s / 15. + rm * 4.

        a = Angle(sexagesimal=(h, m, s), latitude=self.latitude)
        a.negative = self.negative

        return a

    def toround(self, rval=3):

        # Decompose angle
        d, m, s = self.angle

        # Round numbers:
        # 1: degrees only
        # 2: degrees and minutes
        # 3: degrees, minutes, and seconds
        # 4.n: degrees, minutes, and decimal seconds with n decimal places

        if rval < 2:
            n = int(round((rval - 1.) * 100))
            d = round(d + m / 60. + s / 3600., n)
            if n == 0:
                d = int(d)
            return d
        elif rval < 3:
            m = int(round(m + s / 60.))
            if m == 60:
                m = 0
                d = d + 1
            return (d, m)
        elif rval < 4:
            s = int(round(s))
            if s == 60:
                s = 0
                m = m + 1
            if m == 60:
                m = 0
                d = d + 1
            return (d, m, s)
        else:
            n = int(round((rval - 4.) * 100))
            s = round(s, n)
            if s == 60.:
                s = 0.
                m = m + 1
            if m == 60:
                m = 0
                d = d + 1
            return (d, m, s)

    def tostringlist(self, format='ddd:mm:ss', sep=("d", "m", "s")):

        format = format.replace('h', 'd')

        r = 1
        if '.d' in format:
            r = 1
            pos = format.find('.')
            nd = len(format[pos + 1:])
            r = r + nd / 100.
        if 'mm' in format:
            r = 2
        if 'ss' in format:
            r = 3
        if '.s' in format:
            r = 4
            pos = format.find('.')
            ns = len(format[pos + 1:])
            r = r + ns / 100.

        tup = self.toround(rval=r)
        if type(tup) == tuple:
            tup = list(tup)
        else:
            tup = [tup]

        string = []

        if 'dd' in format:
            if '.d' in format:
                string.append(("%0" + str(nd + 3) + "." + str(nd) + "f") % \
                    tup[0] + sep[0])
            else:
                string.append("%i" % tup[0] + sep[0])
        if 'mm' in format:
            string.append("%02i" % tup[1] + sep[1])
        if 'ss' in format and not '.s' in format:
            string.append("%02i" % tup[2] + sep[2])
        if 'ss.s' in format:
            string.append(("%0" + str(ns + 3) + "." + str(ns) + "f") % tup[2] + sep[2])

        # If style is colons, need to remove trailing colon
        if len(string) >= 1 and sep[0] == ':' and not 'mm' in format:
            string[0] = string[0][:-1]
        if len(string) >= 2 and sep[1] == ':' and not 'ss' in format:
            string[1] = string[1][:-1]

        if self.latitude:
            if self.negative:
                string[0] = "-" + string[0]
            else:
                string[0] = "+" + string[0]

        return string

    def __str__(self):
        return self.angle.__str__()

    def __repr__(self):
        return self.angle.__repr__()

    def __add__(self, other):

        s = self.angle[2] + other.angle[2]
        m = self.angle[1] + other.angle[1]
        d = self.angle[0] + other.angle[0]

        s = Angle(sexagesimal=(d, m, s), latitude=self.latitude)
        s._simplify()

        return s

    def __mul__(self, other):

        d, m, s = self.angle
        s = s * other
        m = m * other
        d = d * other

        if self.latitude and self.negative:
            d, m, s = -d, -m, -s

        s = Angle(sexagesimal=(d, m, s), latitude=self.latitude)
        s._simplify()

        return s

    def __eq__(self, other):
        return self.angle[0] == other.angle[0] \
           and self.angle[1] == other.angle[1] \
           and almost_equal(self.angle[2], other.angle[2])

    def __div__(self, other):
        '''
        Divide an angle by another

        This method calculates the division using the angles in degrees, and
        then corrects for any rounding errors if the division should be exact.
        '''

        # Find division of angles in degrees
        div = self.todegrees() / other.todegrees()

        # Find the nearest integer
        divint = int(round(div))

        # Check whether the denominator multiplied by this number is exactly
        # the numerator
        if other * divint == self:
            return divint
        else:
            return div

    __truediv__ = __div__

def smart_round_angle_sexagesimal(x, latitude=False, hours=False):

    d, m, s = 0, 0, 0.

    divisors_360 = math_util.divisors(360)
    divisors_10 = math_util.divisors(10)
    divisors_60 = math_util.divisors(60)

    if hours:
        x /= 15.

    if x >= 1:
        d = math_util.closest(divisors_360, x)
    else:
        x = x * 60.
        if x >= 1:
            m = math_util.closest(divisors_60, x)
        else:
            x = x * 60.
            if x >= 1:
                s = math_util.closest(divisors_60, x)
            else:
                t = 1.
                while True:
                    t = t * 10.
                    x = x * 10.
                    if x >= 1:
                        s = math_util.closest(divisors_10, x) / t
                        break

    a = Angle(sexagesimal=(d, m, s), latitude=latitude)

    if hours:
        a *= 15

    return a


def smart_round_angle_decimal(x, latitude=False):

    divisors_360 = math_util.divisors(360)
    divisors_10 = math_util.divisors(10)

    if x >= 1:
        d = math_util.closest(divisors_360, x)
    else:
        t = 1.
        while True:
            t = t * 10.
            x = x * 10.
            if x >= 1:
                d = math_util.closest(divisors_10, x) / t
                break

    a = Angle(degrees=d, latitude=latitude)

    return a


def _get_label_precision(format, latitude=False):

    # Find base spacing
    if "mm" in format:
        if "ss" in format:
            if "ss.s" in format:
                n_decimal = len(format.split('.')[1])
                label_spacing = Angle(sexagesimal=(0, 0, 10 ** (-n_decimal)), latitude=latitude)
            else:
                label_spacing = Angle(sexagesimal=(0, 0, 1), latitude=latitude)
        else:
            label_spacing = Angle(sexagesimal=(0, 1, 0), latitude=latitude)
    elif "." in format:
        ns = len(format.split('.')[1])
        label_spacing = Angle(degrees=10 ** (-ns), latitude=latitude)
    else:
        label_spacing = Angle(sexagesimal=(1, 0, 0), latitude=latitude)

    # Check if hours are used instead of degrees
    if "hh" in format:
        label_spacing *= 15

    return label_spacing


class InconsistentSpacing(Exception):
    pass


def _check_format_spacing_consistency(format, spacing):
    '''
    Check whether the format can correctly show labels with the specified
    spacing.

    For example, if the tick spacing is set to 1 arcsecond, but the format is
    set to dd:mm, then the labels cannot be correctly shown. Similarly, if the
    spacing is set to 1/1000 of a degree, or 3.6", then a format of dd:mm:ss
    will cause rounding errors, because the spacing includes fractional
    arcseconds.

    This function will raise a warning if the format and spacing are
    inconsistent.
    '''

    label_spacing = _get_label_precision(format)

    if type(spacing / label_spacing) != int:
        raise InconsistentSpacing('Label format and tick spacing are inconsistent. Make sure that the tick spacing is a multiple of the smallest angle that can be represented by the specified format (currently %s). For example, if the format is dd:mm:ss.s, then the tick spacing has to be a multiple of 0.1". Similarly, if the format is hh:mm:ss, then the tick spacing has to be a multiple of 15". If you got this error as a result of interactively zooming in to a small region, this means that the default display format for the labels is not accurate enough, so you will need to increase the format precision.' % format)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
APLpy : Astronomical Plotting Library in Python
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:

    from .core import FITSFigure
    from .rgb import make_rgb_image, make_rgb_cube

    from .frame import Frame
    from .overlays import Scalebar, Beam
    from .colorbar import Colorbar
    from .grid import Grid
    from .ticks import Ticks
    from .labels import TickLabels
    from .axis_labels import AxisLabels



from __future__ import absolute_import, print_function, division

import warnings

from .decorators import auto_refresh


class Deprecated(object):

    @auto_refresh
    def set_tick_xspacing(self, *args, **kwargs):
        warnings.warn("set_tick_xspacing is deprecated - use ticks.set_xspacing instead", DeprecationWarning)
        self.ticks.set_xspacing(*args, **kwargs)

    @auto_refresh
    def set_tick_yspacing(self, *args, **kwargs):
        warnings.warn("set_tick_yspacing is deprecated - use ticks.set_yspacing instead", DeprecationWarning)
        self.ticks.set_yspacing(*args, **kwargs)

    @auto_refresh
    def set_tick_size(self, *args, **kwargs):
        warnings.warn("set_tick_size is deprecated - use ticks.set_length instead", DeprecationWarning)
        self.ticks.set_length(*args, **kwargs)

    @auto_refresh
    def set_tick_color(self, *args, **kwargs):
        warnings.warn("set_tick_color is deprecated - use ticks.set_color instead", DeprecationWarning)
        self.ticks.set_color(*args, **kwargs)

    @auto_refresh
    def show_grid(self, *args, **kwargs):
        warnings.warn("show_grid is deprecated - use add_grid instead", DeprecationWarning)
        self.add_grid(*args, **kwargs)

    @auto_refresh
    def hide_grid(self, *args, **kwargs):
        warnings.warn("hide_grid is deprecated - use remove_grid instead", DeprecationWarning)
        self.remove_grid(*args, **kwargs)

    @auto_refresh
    def show_beam(self, *args, **kwargs):
        warnings.warn("show_beam is deprecated - use add_beam instead", DeprecationWarning)
        self.add_beam(*args, **kwargs)

    @auto_refresh
    def hide_beam(self, *args, **kwargs):
        warnings.warn("hide_beam is deprecated - use remove_beam instead", DeprecationWarning)
        self.add_beam(*args, **kwargs)

    @auto_refresh
    def show_scalebar(self, *args, **kwargs):
        warnings.warn("show_scalebar is deprecated - use add_scalebar instead", DeprecationWarning)
        self.add_scalebar(*args, **kwargs)

    @auto_refresh
    def hide_scalebar(self, *args, **kwargs):
        warnings.warn("hide_scalebar is deprecated - use remove_scalebar instead", DeprecationWarning)
        self.add_scalebar(*args, **kwargs)

    @auto_refresh
    def show_colorbar(self, *args, **kwargs):
        warnings.warn("show_colorbar is deprecated - use add_colorbar instead", DeprecationWarning)
        self.add_colorbar(*args, **kwargs)

    @auto_refresh
    def hide_colorbar(self, *args, **kwargs):
        warnings.warn("hide_colorbar is deprecated - use remove_colorbar instead", DeprecationWarning)
        self.add_colorbar(*args, **kwargs)

    @auto_refresh
    def set_grid_alpha(self, *args, **kwargs):
        warnings.warn("set_grid_alpha is deprecated - use grid.set_alpha instead", DeprecationWarning)
        self.grid.set_alpha(*args, **kwargs)

    @auto_refresh
    def set_grid_color(self, *args, **kwargs):
        warnings.warn("set_grid_color is deprecated - use grid.set_color instead", DeprecationWarning)
        self.grid.set_color(*args, **kwargs)

    @auto_refresh
    def set_grid_xspacing(self, *args, **kwargs):
        warnings.warn("set_grid_xspacing is deprecated - use grid.set_xspacing instead", DeprecationWarning)
        self.grid.set_xspacing(*args, **kwargs)

    @auto_refresh
    def set_grid_yspacing(self, *args, **kwargs):
        warnings.warn("set_grid_yspacing is deprecated - use grid.set_yspacing instead", DeprecationWarning)
        self.grid.set_yspacing(*args, **kwargs)

    @auto_refresh
    def set_scalebar_properties(self, *args, **kwargs):
        warnings.warn("set_scalebar_properties is deprecated - use scalebar.set instead", DeprecationWarning)
        self.scalebar._set_scalebar_properties(*args, **kwargs)

    @auto_refresh
    def set_label_properties(self, *args, **kwargs):
        warnings.warn("set_label_properties is deprecated - use scalebar.set instead", DeprecationWarning)
        self.scalebar._set_label_properties(*args, **kwargs)

    @auto_refresh
    def set_beam_properties(self, *args, **kwargs):
        warnings.warn("set_beam_properties is deprecated - use beam.set instead", DeprecationWarning)
        self.beam.set(*args, **kwargs)

    @auto_refresh
    def set_labels_latex(self, usetex):
        warnings.warn("set_labels_latex has been deprecated - use set_system_latex instead", DeprecationWarning)
        self.set_system_latex(usetex)

    # TICK LABELS

    @auto_refresh
    def set_tick_labels_format(self, xformat=None, yformat=None):
        warnings.warn("set_tick_labels_format has been deprecated - use tick_labels.set_xformat() and tick_labels.set_yformat instead", DeprecationWarning)
        if xformat:
            self.tick_labels.set_xformat(xformat)
        if yformat:
            self.tick_labels.set_yformat(yformat)

    @auto_refresh
    def set_tick_labels_xformat(self, format):
        warnings.warn("set_tick_labels_xformat has been deprecated - use tick_labels.set_xformat() instead", DeprecationWarning)
        self.tick_labels.set_xformat(format)

    @auto_refresh
    def set_tick_labels_yformat(self, format):
        warnings.warn("set_tick_labels_yformat has been deprecated - use tick_labels.set_yformat() instead", DeprecationWarning)
        self.tick_labels.set_yformat(format)

    @auto_refresh
    def set_tick_labels_style(self, style):
        warnings.warn("set_tick_labels_style has been deprecated - use tick_labels.set_style instead", DeprecationWarning)
        self.tick_labels.set_style(style)

    @auto_refresh
    def set_tick_labels_size(self, size):
        warnings.warn("set_tick_labels_size has been deprecated - use tick_labels.set_font instead", DeprecationWarning)
        self.tick_labels.set_font(size=size)

    @auto_refresh
    def set_tick_labels_weight(self, weight):
        warnings.warn("set_tick_labels_weight has been deprecated - use tick_labels.set_font instead", DeprecationWarning)
        self.tick_labels.set_font(weight=weight)

    @auto_refresh
    def set_tick_labels_family(self, family):
        warnings.warn("set_tick_labels_family has been deprecated - use tick_labels.set_font instead", DeprecationWarning)
        self.tick_labels.set_font(family=family)

    @auto_refresh
    def set_tick_labels_font(self, *args, **kwargs):
        warnings.warn("set_tick_labels_font has been deprecated - use tick_labels.set_font instead", DeprecationWarning)
        self.tick_labels.set_font(*args, **kwargs)

    @auto_refresh
    def show_tick_labels(self):
        warnings.warn("show_tick_labels has been deprecated - use tick_labels.show instead", DeprecationWarning)
        self.tick_labels.show()

    @auto_refresh
    def hide_tick_labels(self):
        warnings.warn("hide_tick_labels has been deprecated - use tick_labels.hide instead", DeprecationWarning)
        self.tick_labels.hide()

    @auto_refresh
    def show_xtick_labels(self):
        warnings.warn("show_xtick_labels has been deprecated - use tick_labels.show_x instead", DeprecationWarning)
        self.tick_labels.show_x()

    @auto_refresh
    def hide_xtick_labels(self):
        warnings.warn("hide_xtick_labels has been deprecated - use tick_labels.hide_x instead", DeprecationWarning)
        self.tick_labels.hide_x()

    @auto_refresh
    def show_ytick_labels(self):
        warnings.warn("show_ytick_labels has been deprecated - use tick_labels.show_y instead", DeprecationWarning)
        self.tick_labels.show_y()

    @auto_refresh
    def hide_ytick_labels(self):
        warnings.warn("hide_ytick_labels has been deprecated - use tick_labels.hide_y instead", DeprecationWarning)
        self.tick_labels.hide_y()

    # AXIS LABELS

    @auto_refresh
    def set_axis_labels(self, xlabel='default', ylabel='default', xpad='default', ypad='default'):
        warnings.warn("set_axis_labels has been deprecated - use axis_labels.set_xtext, axis_labels.set_ytext, axis_labels.set_xpad, and axis_labels.set_ypad instead", DeprecationWarning)
        if xlabel != 'default':
            self.axis_labels.set_xtext(xlabel)
        if xpad != 'default':
            self.axis_labels.set_xpad(xpad)
        if ylabel != 'default':
            self.axis_labels.set_ytext(ylabel)
        if ypad != 'default':
            self.axis_labels.set_ypad(ypad)

    @auto_refresh
    def set_axis_labels_xdisp(self, xpad):
        warnings.warn("set_axis_labels_xdisk has been deprecated - use axis_labels.set_xpad instead", DeprecationWarning)
        self.axis_labels.set_xpad(xpad)

    @auto_refresh
    def set_axis_labels_ydisp(self, ypad):
        warnings.warn("set_axis_labels_xdisp has been deprecated - use axis_labels.set_ypad instead", DeprecationWarning)
        self.axis_labels.set_ypad(ypad)

    @auto_refresh
    def set_axis_labels_size(self, size):
        warnings.warn("set_axis_labels_size has been deprecated - use axis_labels.set_font instead", DeprecationWarning)
        self.axis_labels.set_font(size=size)

    @auto_refresh
    def set_axis_labels_weight(self, weight):
        warnings.warn("set_axis_labels_weight has been deprecated - use axis_labels.set_font instead", DeprecationWarning)
        self.axis_labels.set_font(weight=weight)

    @auto_refresh
    def set_axis_labels_family(self, family):
        warnings.warn("set_axis_labels_family has been deprecated - use axis_labels.set_font instead", DeprecationWarning)
        self.axis_labels.set_font(family=family)

    @auto_refresh
    def set_axis_labels_font(self, *args, **kwargs):
        warnings.warn("set_axis_labels_font has been deprecated - use axis_labels.set_font instead", DeprecationWarning)
        self.axis_labels.set_font(*args, **kwargs)

    @auto_refresh
    def show_axis_labels(self):
        warnings.warn("show_axis_labels has been deprecated - use axis_labels.show instead", DeprecationWarning)
        self.axis_labels.show()

    @auto_refresh
    def hide_axis_labels(self):
        warnings.warn("hide_axis_labels has been deprecated - use axis_labels.hide instead", DeprecationWarning)
        self.axis_labels.hide()

    @auto_refresh
    def show_xaxis_label(self):
        warnings.warn("show_xaxis_label has been deprecated - use axis_labels.show_x instead", DeprecationWarning)
        self.axis_labels.show_x()

    @auto_refresh
    def hide_xaxis_label(self):
        warnings.warn("hide_xaxis_label has been deprecated - use axis_labels.hide_x instead", DeprecationWarning)
        self.axis_labels.hide_x()

    @auto_refresh
    def show_yaxis_label(self):
        warnings.warn("show_yaxis_label has been deprecated - use axis_labels.show_y instead", DeprecationWarning)
        self.axis_labels.show_y()

    @auto_refresh
    def hide_yaxis_label(self):
        warnings.warn("hide_yaxis_label has been deprecated - use axis_labels.hide_y instead", DeprecationWarning)
        self.axis_labels.hide_y()

    # FRAME

    @auto_refresh
    def set_frame_color(self, *args, **kwargs):
        warnings.warn("set_frame_color has been deprecated - use frame.set_color instead", DeprecationWarning)
        self.frame.set_color(*args, **kwargs)

    @auto_refresh
    def set_frame_linewidth(self, *args, **kwargs):
        warnings.warn("set_frame_linewidth has been deprecated - use frame.set_linewidth instead", DeprecationWarning)
        self.frame.set_linewidth(*args, **kwargs)



from __future__ import absolute_import, print_function, division

import numpy as np
from astropy import log

from . import math_util as m


class interp1d(object):

    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.dy = np.zeros(y.shape, dtype=y.dtype)
        self.dy[:-1] = (self.y[1:] - self.y[:-1]) / (self.x[1:] - self.x[:-1])
        self.dy[-1] = self.dy[-2]

    def __call__(self, x_new):

        ipos = np.searchsorted(self.x, x_new)

        if m.isnumeric(x_new):
            if ipos == 0:
                ipos = 1
            if ipos == len(self.x):
                ipos = len(self.x) - 1
        else:
            ipos[ipos == 0] = 1
            ipos[ipos == len(self.x)] = len(self.x) - 1

        ipos = ipos - 1

        return (x_new - self.x[ipos]) * self.dy[ipos] + self.y[ipos]


def resample(array, factor):

    nx, ny = np.shape(array)

    nx_new = nx // factor
    ny_new = ny // factor

    array2 = np.zeros((nx_new, ny))
    for i in range(nx_new):
        array2[i, :] = np.mean(array[i * factor:(i + 1) * factor, :], axis=0)

    array3 = np.zeros((nx_new, ny_new))
    for j in range(ny_new):
        array3[:, j] = np.mean(array2[:, j * factor:(j + 1) * factor], axis=1)

    return array3


def percentile_function(array):

    if np.all(np.isnan(array) | np.isinf(array)):
        log.warning("Image contains only NaN or Inf values")
        return lambda x: 0

    array = array.ravel()
    array = array[np.where(np.isnan(array) == False)]
    array = array[np.where(np.isinf(array) == False)]

    n_total = np.shape(array)[0]

    if n_total == 0:
        def return_zero(x):
            return 0
        return return_zero
    elif n_total == 1:
        def return_single(x):
            return array[0]
        return return_single

    array = np.sort(array)

    x = np.linspace(0., 100., num=n_total)

    spl = interp1d(x=x, y=array)

    if n_total > 10000:
        x = np.linspace(0., 100., num=10000)
        spl = interp1d(x=x, y=spl(x))

    array = None

    return spl


def stretch(array, function, exponent=2, midpoint=None):

    if function == 'linear':
        return array
    elif function == 'log':
        if not m.isnumeric(midpoint):
            midpoint = 0.05
        return np.log10(array / midpoint + 1.) / np.log10(1. / midpoint + 1.)
    elif function == 'sqrt':
        return np.sqrt(array)
    elif function == 'arcsinh':
        if not m.isnumeric(midpoint):
            midpoint = -0.033
        return np.arcsinh(array / midpoint) / np.arcsinh(1. / midpoint)
    elif function == 'power':
        return np.power(array, exponent)
    else:
        raise Exception("Unknown function : " + function)


def _matplotlib_pil_bug_present():
    """
    Determine whether PIL images should be pre-flipped due to a bug in Matplotlib.

    Prior to Matplotlib 1.2.0, RGB images provided as PIL objects were
    oriented wrongly. This function tests whether the bug is present.
    """

    from matplotlib.image import pil_to_array

    try:
        from PIL import Image
    except:
        import Image

    from astropy import log

    array1 = np.array([[1,2],[3,4]], dtype=np.uint8)
    image = Image.fromarray(array1)
    array2 = pil_to_array(image)

    if np.all(array1 == array2):
        log.debug("PIL Image flipping bug not present in Matplotlib")
        return False
    elif np.all(array1 == array2[::-1,:]):
        log.debug("PIL Image flipping bug detected in Matplotlib")
        return True
    else:
        log.warning("Could not properly determine Matplotlib behavior for RGB images - image may be flipped incorrectly")
        return False



from __future__ import absolute_import, print_function, division

from distutils import version
import os
import warnings

import tempfile
import shutil

import numpy as np
from astropy.extern import six
from astropy import log
from astropy.io import fits

from . import image_util
from . import math_util


def _data_stretch(image, vmin=None, vmax=None, pmin=0.25, pmax=99.75,
                  stretch='linear', vmid=None, exponent=2):

    min_auto = not math_util.isnumeric(vmin)
    max_auto = not math_util.isnumeric(vmax)

    if min_auto or max_auto:
        auto_v = image_util.percentile_function(image)
        vmin_auto, vmax_auto = auto_v(pmin), auto_v(pmax)

    if min_auto:
        log.info("vmin = %10.3e (auto)" % vmin_auto)
        vmin = vmin_auto
    else:
        log.info("vmin = %10.3e" % vmin)

    if max_auto:
        log.info("vmax = %10.3e (auto)" % vmax_auto)
        vmax = vmax_auto
    else:
        log.info("vmax = %10.3e" % vmax)

    image = (image - vmin) / (vmax - vmin)

    data = image_util.stretch(image, stretch, exponent=exponent, midpoint=vmid)

    data = np.nan_to_num(data)
    data = np.clip(data * 255., 0., 255.)

    return data.astype(np.uint8)


def make_rgb_image(data, output, indices=(0, 1, 2), \
                   vmin_r=None, vmax_r=None, pmin_r=0.25, pmax_r=99.75, \
                   stretch_r='linear', vmid_r=None, exponent_r=2, \
                   vmin_g=None, vmax_g=None, pmin_g=0.25, pmax_g=99.75, \
                   stretch_g='linear', vmid_g=None, exponent_g=2, \
                   vmin_b=None, vmax_b=None, pmin_b=0.25, pmax_b=99.75, \
                   stretch_b='linear', vmid_b=None, exponent_b=2, \
                   make_nans_transparent=False, \
                   embed_avm_tags=True):
    '''
    Make an RGB image from a FITS RGB cube or from three FITS files.

    Parameters
    ----------

    data : str or tuple or list
        If a string, this is the filename of an RGB FITS cube. If a tuple
        or list, this should give the filename of three files to use for
        the red, green, and blue channel.

    output : str
        The output filename. The image type (e.g. PNG, JPEG, TIFF, ...)
        will be determined from the extension. Any image type supported by
        the Python Imaging Library can be used.

    indices : tuple, optional
        If data is the filename of a FITS cube, these indices are the
        positions in the third dimension to use for red, green, and
        blue respectively. The default is to use the first three
        indices.

    vmin_r, vmin_g, vmin_b : float, optional
        Minimum pixel value to use for the red, green, and blue channels.
        If set to None for a given channel, the minimum pixel value for
        that channel is determined using the corresponding pmin_x argument
        (default).

    vmax_r, vmax_g, vmax_b : float, optional
        Maximum pixel value to use for the red, green, and blue channels.
        If set to None for a given channel, the maximum pixel value for
        that channel is determined using the corresponding pmax_x argument
        (default).

    pmin_r, pmin_r, pmin_g : float, optional
        Percentile values used to determine for a given channel the
        minimum pixel value to use for that channel if the corresponding
        vmin_x is set to None. The default is 0.25% for all channels.

    pmax_r, pmax_g, pmax_b : float, optional
        Percentile values used to determine for a given channel the
        maximum pixel value to use for that channel if the corresponding
        vmax_x is set to None. The default is 99.75% for all channels.

    stretch_r, stretch_g, stretch_b : { 'linear', 'log', 'sqrt', 'arcsinh', 'power' }
        The stretch function to use for the different channels.

    vmid_r, vmid_g, vmid_b : float, optional
        Baseline values used for the log and arcsinh stretches. If
        set to None, this is set to zero for log stretches and to
        vmin - (vmax - vmin) / 30. for arcsinh stretches

    exponent_r, exponent_g, exponent_b : float, optional
        If stretch_x is set to 'power', this is the exponent to use.

    make_nans_transparent : bool, optional
        If set AND output is png, will add an alpha layer that sets pixels
        containing a NaN to transparent.

    embed_avm_tags : bool, optional
        Whether to embed AVM tags inside the image - this can only be done for
        JPEG and PNG files, and only if PyAVM is installed.
    '''

    try:
        from PIL import Image
    except ImportError:
        try:
            import Image
        except ImportError:
            raise ImportError("The Python Imaging Library (PIL) is required to make an RGB image")

    if isinstance(data, six.string_types):

        image = fits.getdata(data)
        image_r = image[indices[0], :, :]
        image_g = image[indices[1], :, :]
        image_b = image[indices[2], :, :]

        # Read in header
        header = fits.getheader(data)

        # Remove information about third dimension
        header['NAXIS'] = 2
        for key in ['NAXIS', 'CTYPE', 'CRPIX', 'CRVAL', 'CUNIT', 'CDELT', 'CROTA']:
            for coord in range(3, 6):
                name = key + str(coord)
                if name in header:
                    header.__delitem__(name)

    elif (type(data) == list or type(data) == tuple) and len(data) == 3:

        filename_r, filename_g, filename_b = data
        image_r = fits.getdata(filename_r)
        image_g = fits.getdata(filename_g)
        image_b = fits.getdata(filename_b)

        # Read in header
        header = fits.getheader(filename_r)

    else:
        raise Exception("data should either be the filename of a FITS cube or a list/tuple of three images")

    # are we making a transparent layer?
    do_alpha = make_nans_transparent and output.lower().endswith('.png')

    if do_alpha:
        log.info("Making alpha layer")

        # initialize alpha layer
        image_alpha = np.empty_like(image_r, dtype=np.uint8)
        image_alpha[:] = 255

        # look for nans in images
        for im in [image_r, image_g, image_b]:
            image_alpha[np.isnan(im)] = 0

    log.info("Red:")
    image_r = Image.fromarray(_data_stretch(image_r, \
                                            vmin=vmin_r, vmax=vmax_r, \
                                            pmin=pmin_r, pmax=pmax_r, \
                                            stretch=stretch_r, \
                                            vmid=vmid_r, \
                                            exponent=exponent_r))

    log.info("Green:")
    image_g = Image.fromarray(_data_stretch(image_g, \
                                            vmin=vmin_g, vmax=vmax_g, \
                                            pmin=pmin_g, pmax=pmax_g, \
                                            stretch=stretch_g, \
                                            vmid=vmid_g, \
                                            exponent=exponent_g))

    log.info("Blue:")
    image_b = Image.fromarray(_data_stretch(image_b, \
                                            vmin=vmin_b, vmax=vmax_b, \
                                            pmin=pmin_b, pmax=pmax_b, \
                                            stretch=stretch_b, \
                                            vmid=vmid_b, \
                                            exponent=exponent_b))

    img = Image.merge("RGB", (image_r, image_g, image_b))

    if do_alpha:
        # convert to RGBA and add alpha layer
        image_alpha = Image.fromarray(image_alpha)
        img.convert("RGBA")
        img.putalpha(image_alpha)

    img = img.transpose(Image.FLIP_TOP_BOTTOM)

    img.save(output)

    if embed_avm_tags:

        try:
            import pyavm
        except ImportError:
            warnings.warn("PyAVM 0.9.1 or later is not installed, so AVM tags will not be embedded in RGB image")
            return

        if version.LooseVersion(pyavm.__version__) < version.LooseVersion('0.9.1'):
            warnings.warn("PyAVM 0.9.1 or later is not installed, so AVM tags will not be embedded in RGB image")
            return

        from pyavm import AVM

        if output.lower().endswith(('.jpg', '.jpeg', '.png')):
            avm = AVM.from_header(header)
            avm.embed(output, output)
        else:
            warnings.warn("AVM tags will not be embedded in RGB image, as only JPEG and PNG files are supported")


def make_rgb_cube(files, output, north=False, system=None, equinox=None):
    '''
    Make an RGB data cube from a list of three FITS images.

    This method can read in three FITS files with different
    projections/sizes/resolutions and uses Montage to reproject
    them all to the same projection.

    Two files are produced by this function. The first is a three-dimensional
    FITS cube with a filename give by `output`, where the third dimension
    contains the different channels. The second is a two-dimensional FITS
    image with a filename given by `output` with a `_2d` suffix. This file
    contains the mean of the different channels, and is required as input to
    FITSFigure if show_rgb is subsequently used to show a color image
    generated from the FITS cube (to provide the correct WCS information to
    FITSFigure).

    Parameters
    ----------

    files : tuple or list
       A list of the filenames of three FITS filename to reproject.
       The order is red, green, blue.

    output : str
       The filename of the output RGB FITS cube.

    north : bool, optional
       By default, the FITS header generated by Montage represents the
       best fit to the images, often resulting in a slight rotation. If
       you want north to be straight up in your final mosaic, you should
       use this option.

    system : str, optional
       Specifies the system for the header (default is EQUJ).
       Possible values are: EQUJ EQUB ECLJ ECLB GAL SGAL

    equinox : str, optional
       If a coordinate system is specified, the equinox can also be given
       in the form YYYY. Default is J2000.
    '''

    # Check whether the Python montage module is installed. The Python module
    # checks itself whether the Montage command-line tools are available, and
    # if they are not then importing the Python module will fail.
    try:
        import montage_wrapper as montage
    except ImportError:
        raise Exception("Both the Montage command-line tools and the"
                        " montage-wrapper Python module are required"
                        " for this function")

    # Check that input files exist
    for f in files:
        if not os.path.exists(f):
            raise Exception("File does not exist : " + f)

    # Create work directory
    work_dir = tempfile.mkdtemp()

    raw_dir = '%s/raw' % work_dir
    final_dir = '%s/final' % work_dir

    images_raw_tbl = '%s/images_raw.tbl' % work_dir
    header_hdr = '%s/header.hdr' % work_dir

    # Create raw and final directory in work directory
    os.mkdir(raw_dir)
    os.mkdir(final_dir)

    # Create symbolic links to input files
    for i, f in enumerate(files):
        os.symlink(os.path.abspath(f), '%s/image_%i.fits' % (raw_dir, i))

    # List files and create optimal header
    montage.mImgtbl(raw_dir, images_raw_tbl, corners=True)
    montage.mMakeHdr(images_raw_tbl, header_hdr, north_aligned=north, system=system, equinox=equinox)

    # Read header in with astropy.io.fits
    header = fits.Header.fromtextfile(header_hdr)

    # Find image dimensions
    nx = int(header['NAXIS1'])
    ny = int(header['NAXIS2'])

    # Generate empty datacube
    image_cube = np.zeros((len(files), ny, nx), dtype=np.float32)

    # Loop through files
    for i in range(len(files)):

        # Reproject channel to optimal header
        montage.reproject('%s/image_%i.fits' % (raw_dir, i),
                          '%s/image_%i.fits' % (final_dir, i),
                          header=header_hdr, exact_size=True, bitpix=-32)

        # Read in and add to datacube
        image_cube[i, :, :] = fits.getdata('%s/image_%i.fits' % (final_dir, i))

    # Write out final cube
    fits.writeto(output, image_cube, header, clobber=True)

    # Write out collapsed version of cube
    fits.writeto(output.replace('.fits', '_2d.fits'), \
                   np.mean(image_cube, axis=0), header, clobber=True)

    # Remove work directory
    shutil.rmtree(work_dir)



from __future__ import absolute_import, print_function, division

import numpy as np


import warnings
from astropy import log
from astropy.wcs import WCS as AstropyWCS
import astropy.wcs


def decode_ascii(string):
    try:
        return string.decode('ascii')
    except AttributeError:
        return string


class WCS(AstropyWCS):

    def __init__(self, *args, **kwargs):

        if 'slices' in kwargs:
            self._slices = kwargs.pop('slices')
        else:
            self._slices = []

        if 'dimensions' in kwargs:
            self._dimensions = kwargs.pop('dimensions')
        else:
            self._dimensions = [0, 1]

        AstropyWCS.__init__(self, *args, **kwargs)

        # Fix common non-standard units
        self.wcs.unitfix()

        # Now find the values of the coordinates in the slices - only needed if
        # data has more than two dimensions
        if len(self._slices) > 0:

            self.nx = args[0]['NAXIS%i' % (self._dimensions[0] + 1)]
            self.ny = args[0]['NAXIS%i' % (self._dimensions[1] + 1)]
            xpix = np.arange(self.nx) + 1.
            ypix = np.arange(self.ny) + 1.
            xpix, ypix = np.meshgrid(xpix, ypix)
            xpix, ypix = xpix.reshape(self.nx * self.ny), ypix.reshape(self.nx * self.ny)
            s = 0
            coords = []
            for dim in range(self.naxis):
                if dim == self._dimensions[0]:
                    coords.append(xpix)
                elif dim == self._dimensions[1]:
                    coords.append(ypix)
                else:
                    coords.append(np.repeat(self._slices[s], xpix.shape))
                    s += 1
            coords = np.vstack(coords).transpose()
            result = AstropyWCS.wcs_pix2world(self, coords, 1)
            self._mean_world = np.mean(result, axis=0)
            # result = result.transpose()
            # result = result.reshape((result.shape[0],) + (self.ny, self.nx))

        # Now guess what 'type' of values are on each axis
        if self.ctype_x[:4] == 'RA--' or \
           self.ctype_x[1:4] == 'LON':
            self.set_xaxis_coord_type('longitude')
            self.set_xaxis_coord_type('longitude')
        elif self.ctype_x[:4] == 'DEC-' or \
           self.ctype_x[1:4] == 'LAT':
            self.set_xaxis_coord_type('latitude')
            self.set_xaxis_coord_type('latitude')
        else:
            self.set_xaxis_coord_type('scalar')
            self.set_xaxis_coord_type('scalar')

        if self.ctype_y[:4] == 'RA--' or \
           self.ctype_y[1:4] == 'LON':
            self.set_yaxis_coord_type('longitude')
            self.set_yaxis_coord_type('longitude')
        elif self.ctype_y[:4] == 'DEC-' or \
           self.ctype_y[1:4] == 'LAT':
            self.set_yaxis_coord_type('latitude')
            self.set_yaxis_coord_type('latitude')
        else:
            self.set_yaxis_coord_type('scalar')
            self.set_yaxis_coord_type('scalar')

    @property
    def is_celestial(self):
        return self.wcs.lng in self._dimensions and self.wcs.lat in self._dimensions

    @property
    def pixel_scale_matrix(self):

        # Only for the subset of dimensions used here

        d0 = self._dimensions[0]
        d1 = self._dimensions[1]

        cdelt = np.matrix([[self.wcs.get_cdelt()[d0],0],
                           [0, self.wcs.get_cdelt()[d1]]])

        pc_full = self.wcs.get_pc()
        pc = np.matrix([[pc_full[d0,d0], pc_full[d0,d1]],
                        [pc_full[d1,d0], pc_full[d1,d1]]])

        m = np.array(cdelt * pc)

        return m

    def set_xaxis_coord_type(self, coord_type):
        if coord_type in ['longitude', 'latitude', 'scalar']:
            self.xaxis_coord_type = coord_type
        else:
            raise Exception("coord_type should be one of longitude/latitude/scalar")

    def set_yaxis_coord_type(self, coord_type):
        if coord_type in ['longitude', 'latitude', 'scalar']:
            self.yaxis_coord_type = coord_type
        else:
            raise Exception("coord_type should be one of longitude/latitude/scalar")

    def __getattr__(self, attribute):

        if attribute[-2:] == '_x':
            axis = self._dimensions[0]
        elif attribute[-2:] == '_y':
            axis = self._dimensions[1]
        else:
            raise AttributeError("Attribute %s does not exist" % attribute)

        if attribute[:5] == 'ctype':
            return decode_ascii(self.wcs.ctype[axis])
        elif attribute[:5] == 'cname':
            return decode_ascii(self.wcs.cname[axis])
        elif attribute[:5] == 'cunit':
            return str(self.wcs.cunit[axis])
        elif attribute[:5] == 'crval':
            return decode_ascii(self.wcs.crval[axis])
        elif attribute[:5] == 'crpix':
            return decode_ascii(self.wcs.crpix[axis])
        else:
            raise AttributeError("Attribute %s does not exist" % attribute)

    def wcs_world2pix(self, x, y, origin):
        if self.naxis == 2:
            if self._dimensions[1] < self._dimensions[0]:
                xp, yp = AstropyWCS.wcs_world2pix(self, y, x, origin)
                return yp, xp
            else:
                return AstropyWCS.wcs_world2pix(self, x, y, origin)
        else:
            coords = []
            s = 0
            for dim in range(self.naxis):
                if dim == self._dimensions[0]:
                    coords.append(x)
                elif dim == self._dimensions[1]:
                    coords.append(y)
                else:
                    # The following is an approximation, and will break down if
                    # the world coordinate changes significantly over the slice
                    coords.append(np.repeat(self._mean_world[dim], x.shape))
                    s += 1
            coords = np.vstack(coords).transpose()

            # Due to a bug in pywcs, we need to loop over each coordinate
            # result = AstropyWCS.wcs_world2pix(self, coords, origin)
            result = np.zeros(coords.shape)
            for i in range(result.shape[0]):
                result[i:i + 1, :] = AstropyWCS.wcs_world2pix(self, coords[i:i + 1, :], origin)

            return result[:, self._dimensions[0]], result[:, self._dimensions[1]]

    def wcs_pix2world(self, x, y, origin):
        if self.naxis == 2:
            if self._dimensions[1] < self._dimensions[0]:
                xw, yw = AstropyWCS.wcs_pix2world(self, y, x, origin)
                return yw, xw
            else:
                return AstropyWCS.wcs_pix2world(self, x, y, origin)
        else:
            coords = []
            s = 0
            for dim in range(self.naxis):
                if dim == self._dimensions[0]:
                    coords.append(x)
                elif dim == self._dimensions[1]:
                    coords.append(y)
                else:
                    coords.append(np.repeat(self._slices[s] + 0.5, x.shape))
                    s += 1
            coords = np.vstack(coords).transpose()
            result = AstropyWCS.wcs_pix2world(self, coords, origin)
            return result[:, self._dimensions[0]], result[:, self._dimensions[1]]


def convert_coords(x, y, input, output):

    system_in, equinox_in = input
    system_out, equinox_out = output

    if input == output:
        return x, y

    # Need to take into account inverted coords

    if system_in['name'] == 'galactic' and system_out['name'] == 'equatorial':

        if equinox_out == 'j2000':
            x, y = gal2fk5(x, y)
        elif equinox_out == 'b1950':
            x, y = gal2fk5(x, y)
            x, y = j2000tob1950(x, y)
        else:
            raise Exception("Cannot convert from galactic to equatorial coordinates for equinox=%s" % equinox_out)

    elif system_in['name'] == 'equatorial' and system_out['name'] == 'galactic':

        if equinox_in == 'j2000':
            x, y = fk52gal(x, y)
        elif equinox_in == 'b1950':
            x, y = b1950toj2000(x, y)
            x, y = fk52gal(x, y)
        else:
            raise Exception("Cannot convert from equatorial to equatorial coordinates for equinox=%s" % equinox_in)

    elif system_in['name'] == 'equatorial' and system_out['name'] == 'equatorial':

        if equinox_in == 'b1950' and equinox_out == 'j2000':
            x, y = b1950toj2000(x, y)
        elif equinox_in == 'j2000' and equinox_out == 'b1950':
            x, y = j2000tob1950(x, y)
        elif equinox_in == equinox_out:
            pass
        else:
            raise Exception("Cannot convert between equatorial coordinates for equinoxes %s and %s" % (equinox_in, equinox_out))

    else:
        raise Exception("Cannot (yet) convert between %s, %s and %s, %s" % (system_in['name'], equinox_in, system_out['name'], equinox_out))

    # Take into account inverted coordinates
    if system_in['inverted'] is system_out['inverted']:
        return x, y
    else:
        return y, x


def precession_matrix(equinox1, equinox2, fk4=False):
    "Adapted from the IDL astronomy library"

    deg_to_rad = np.pi / 180.
    sec_to_rad = deg_to_rad / 3600.

    t = 0.001 * (equinox2 - equinox1)

    if not fk4:

        st = 0.001 * (equinox1 - 2000.)

        # Compute 3 rotation angles
        a = sec_to_rad * t * (23062.181 + st * (139.656 + 0.0139 * st) + t * (30.188 - 0.344 * st + 17.998 * t))
        b = sec_to_rad * t * t * (79.280 + 0.410 * st + 0.205 * t) + a
        c = sec_to_rad * t * (20043.109 - st * (85.33 + 0.217 * st) + t * (- 42.665 - 0.217 * st - 41.833 * t))

    else:

        st = 0.001 * (equinox1 - 1900.)

        # Compute 3 rotation angles
        a = sec_to_rad * t * (23042.53 + st * (139.75 + 0.06 * st) + t * (30.23 - 0.27 * st + 18.0 * t))
        b = sec_to_rad * t * t * (79.27 + 0.66 * st + 0.32 * t) + a
        c = sec_to_rad * t * (20046.85 - st * (85.33 + 0.37 * st) + t * (- 42.67 - 0.37 * st - 41.8 * t))

    sina = np.sin(a)
    sinb = np.sin(b)
    sinc = np.sin(c)
    cosa = np.cos(a)
    cosb = np.cos(b)
    cosc = np.cos(c)

    r = np.matrix([[cosa * cosb * cosc - sina * sinb, sina * cosb + cosa * sinb * cosc,  cosa * sinc],
                   [- cosa * sinb - sina * cosb * cosc, cosa * cosb - sina * sinb * cosc, - sina * sinc],
                   [- cosb * sinc, - sinb * sinc, cosc]])

    return r

P1 = precession_matrix(1950., 2000.)
P2 = precession_matrix(2000., 1950.)


def b1950toj2000(ra, dec):
    '''
    Convert B1950 to J2000 coordinates.

    This routine is based on the technique described at
    http://www.stargazing.net/kepler/b1950.html
    '''

    # Convert to radians
    ra = np.radians(ra)
    dec = np.radians(dec)

    # Convert RA, Dec to rectangular coordinates
    x = np.cos(ra) * np.cos(dec)
    y = np.sin(ra) * np.cos(dec)
    z = np.sin(dec)

    # Apply the precession matrix
    x2 = P1[0, 0] * x + P1[1, 0] * y + P1[2, 0] * z
    y2 = P1[0, 1] * x + P1[1, 1] * y + P1[2, 1] * z
    z2 = P1[0, 2] * x + P1[1, 2] * y + P1[2, 2] * z

    # Convert the new rectangular coordinates back to RA, Dec
    ra = np.arctan2(y2, x2)
    dec = np.arcsin(z2)

    # Convert to degrees
    ra = np.degrees(ra)
    dec = np.degrees(dec)

    # Make sure ra is between 0. and 360.
    ra = np.mod(ra, 360.)
    dec = np.mod(dec + 90., 180.) - 90.

    return ra, dec


def j2000tob1950(ra, dec):
    '''
    Convert J2000 to B1950 coordinates.

    This routine was derived by taking the inverse of the b1950toj2000 routine
    '''

    # Convert to radians
    ra = np.radians(ra)
    dec = np.radians(dec)

    # Convert RA, Dec to rectangular coordinates
    x = np.cos(ra) * np.cos(dec)
    y = np.sin(ra) * np.cos(dec)
    z = np.sin(dec)

    # Apply the precession matrix
    x2 = P2[0, 0] * x + P2[1, 0] * y + P2[2, 0] * z
    y2 = P2[0, 1] * x + P2[1, 1] * y + P2[2, 1] * z
    z2 = P2[0, 2] * x + P2[1, 2] * y + P2[2, 2] * z

    # Convert the new rectangular coordinates back to RA, Dec
    ra = np.arctan2(y2, x2)
    dec = np.arcsin(z2)

    # Convert to degrees
    ra = np.degrees(ra)
    dec = np.degrees(dec)

    # Make sure ra is between 0. and 360.
    ra = np.mod(ra, 360.)
    dec = np.mod(dec + 90., 180.) - 90.

    return ra, dec

# Galactic conversion constants
RA_NGP = np.radians(192.859508333333)
DEC_NGP = np.radians(27.1283361111111)
L_CP = np.radians(122.932)
L_0 = L_CP - np.pi / 2.
RA_0 = RA_NGP + np.pi / 2.
DEC_0 = np.pi / 2. - DEC_NGP


def gal2fk5(l, b):

    l = np.radians(l)
    b = np.radians(b)

    sind = np.sin(b) * np.sin(DEC_NGP) + np.cos(b) * np.cos(DEC_NGP) * np.sin(l - L_0)

    dec = np.arcsin(sind)

    cosa = np.cos(l - L_0) * np.cos(b) / np.cos(dec)
    sina = (np.cos(b) * np.sin(DEC_NGP) * np.sin(l - L_0) - np.sin(b) * np.cos(DEC_NGP)) / np.cos(dec)

    dec = np.degrees(dec)

    ra = np.arccos(cosa)
    ra[np.where(sina < 0.)] = -ra[np.where(sina < 0.)]

    ra = np.degrees(ra + RA_0)

    ra = np.mod(ra, 360.)
    dec = np.mod(dec + 90., 180.) - 90.

    return ra, dec


def fk52gal(ra, dec):

    ra, dec = np.radians(ra), np.radians(dec)

    sinb = np.sin(dec) * np.cos(DEC_0) - np.cos(dec) * np.sin(ra - RA_0) * np.sin(DEC_0)

    b = np.arcsin(sinb)

    cosl = np.cos(dec) * np.cos(ra - RA_0) / np.cos(b)
    sinl = (np.sin(dec) * np.sin(DEC_0) + np.cos(dec) * np.sin(ra - RA_0) * np.cos(DEC_0)) / np.cos(b)

    b = np.degrees(b)

    l = np.arccos(cosl)
    l[np.where(sinl < 0.)] = - l[np.where(sinl < 0.)]

    l = np.degrees(l + L_0)

    l = np.mod(l, 360.)
    b = np.mod(b + 90., 180.) - 90.

    return l, b


def system(wcs):

    xcoord = wcs.ctype_x[0:4]
    ycoord = wcs.ctype_y[0:4]
    equinox = wcs.wcs.equinox

    system = {}

    if xcoord == 'RA--' and ycoord == 'DEC-':
        system['name'] = 'equatorial'
        system['inverted'] = False
    elif ycoord == 'RA--' and xcoord == 'DEC-':
        system['name'] = 'equatorial'
        system['inverted'] = True
    elif xcoord == 'GLON' and ycoord == 'GLAT':
        system['name'] = 'galactic'
        system['inverted'] = False
    elif ycoord == 'GLON' and xcoord == 'GLAT':
        system['name'] = 'galactic'
        system['inverted'] = True
    elif xcoord == 'ELON' and ycoord == 'ELAT':
        system['name'] = 'ecliptic'
        system['inverted'] = False
    elif ycoord == 'ELON' and xcoord == 'ELAT':
        system['name'] = 'ecliptic'
        system['inverted'] = True
    else:
        system['name'] = 'unknown'
        system['inverted'] = False

    if system['name'] == 'equatorial':
        if equinox == '' or np.isnan(equinox) or equinox == 0.:
            log.warning("Cannot determine equinox. Assuming J2000.")
            equinox = 'j2000'
        elif equinox == 1950.:
            equinox = 'b1950'
        elif equinox == 2000.:
            equinox = 'j2000'
        else:
            raise Exception("Cannot use equinox %s" % equinox)
    else:
        equinox = 'none'

    units = 'degrees'

    return system, equinox, units


def world2pix(wcs, x_world, y_world):
    if np.isscalar(x_world) and np.isscalar(y_world):
        x_pix, y_pix = wcs.wcs_world2pix(np.array([x_world]), np.array([y_world]), 1)
        return x_pix[0], y_pix[0]
    elif (type(x_world) == list) and (type(y_world) == list):
        x_pix, y_pix = wcs.wcs_world2pix(np.array(x_world), np.array(y_world), 1)
        return x_pix.tolist(), y_pix.tolist()
    elif isinstance(x_world, np.ndarray) and isinstance(y_world, np.ndarray):
        return wcs.wcs_world2pix(x_world, y_world, 1)
    else:
        raise Exception("world2pix should be provided either with two scalars, two lists, or two numpy arrays")


def pix2world(wcs, x_pix, y_pix):
    if np.isscalar(x_pix) and np.isscalar(y_pix):
        x_world, y_world = wcs.wcs_pix2world(np.array([x_pix]), np.array([y_pix]), 1)
        return x_world[0], y_world[0]
    elif (type(x_pix) == list) and (type(y_pix) == list):
        x_world, y_world = wcs.wcs_pix2world(np.array(x_pix), np.array(y_pix), 1)
        return x_world.tolist(), y_world.tolist()
    elif isinstance(x_pix, np.ndarray) and isinstance(y_pix, np.ndarray):
        return wcs.wcs_pix2world(x_pix, y_pix, 1)
    else:
        raise Exception("pix2world should be provided either with two scalars, two lists, or two numpy arrays")


def celestial_pixel_scale(wcs):
    """
    If the pixels are square, return the pixel scale in the spatial
    dimensions
    """

    if not wcs.is_celestial:
        raise ValueError("WCS is not celestial, cannot determine celestial pixel scale")

    # Extract celestial part of WCS and sanitize
    from astropy.wcs import WCSSUB_CELESTIAL
    wcs = wcs.sub([WCSSUB_CELESTIAL])

    pccd = wcs.pixel_scale_matrix

    scale = np.sqrt((pccd ** 2).sum(axis=0))

    if not np.allclose(scale[0], scale[1]):
        warnings.warn("Pixels are not square, using an average pixel scale")
        return np.mean(scale)
    else:
        return scale[0]


def non_celestial_pixel_scales(wcs):

    if wcs.is_celestial:
        raise ValueError("WCS is celestial, use celestial_pixel_scale instead")

    pccd = wcs.pixel_scale_matrix

    if np.allclose(pccd[1,0], 0) and np.allclose(pccd[0,1], 0):
        return np.abs(np.diagonal(pccd))
    else:
        raise ValueError("WCS is rotated, cannot determine consistent pixel scales")



from __future__ import absolute_import, print_function, division


def slice_hypercube(data, header, dimensions=[0, 1], slices=[]):
    '''
    Extract a slice from an n-dimensional HDU data/header pair, and return the
    new data (without changing the header).
    '''

    if type(slices) == int:
        slices = (slices, )
    else:
        slices = slices[:]

    shape = data.shape

    if len(shape) < 2:

        raise Exception("FITS file does not have enough dimensions")

    elif len(shape) == 2:

        if dimensions[1] < dimensions[0]:
            data = data.transpose()

        return data

    else:

        if slices:

            if dimensions[0] < dimensions[1]:
                slices.insert(dimensions[0], slice(None, None, None))
                slices.insert(dimensions[1], slice(None, None, None))
            else:
                slices.insert(dimensions[1], slice(None, None, None))
                slices.insert(dimensions[0], slice(None, None, None))

            if type(slices) == list:
                slices = tuple(slices)

            data = data[slices[::-1]]

            if dimensions[1] < dimensions[0]:
                data = data.transpose()

        else:

            message = '''
    Attempted to read in %i-dimensional FITS cube, but
    dimensions and slices were not specified. Please specify these
    using the dimensions= and slices= argument. The cube dimensions
    are:\n\n''' % len(shape)

            for i in range(1, len(shape) + 1):

                message += " " * 10
                message += " %i %s %i\n" % (i - 1,
                                            header["CTYPE%i" % i],
                                            header["NAXIS%i" % i])

            raise Exception(message)

        return data



from __future__ import absolute_import, print_function, division

import warnings

import numpy as np
from matplotlib.collections import LineCollection

from . import math_util
from . import wcs_util
from . import angle_util as au
from .ticks import tick_positions, default_spacing
from .decorators import auto_refresh


class Grid(object):

    @auto_refresh
    def __init__(self, parent):

        # Save axes and wcs information
        self.ax = parent._ax1
        self._wcs = parent._wcs
        self._figure = parent._figure

        # Save plotting parameters (required for @auto_refresh)
        self._parameters = parent._parameters

        # Initialize grid container
        self._grid = None
        self._active = False

        # Set defaults
        self.x_auto_spacing = True
        self.y_auto_spacing = True
        self.default_color = 'white'
        self.default_alpha = 0.5

        # Set grid event handler
        self.ax.callbacks.connect('xlim_changed', self._update_norefresh)
        self.ax.callbacks.connect('ylim_changed', self._update_norefresh)

    @auto_refresh
    def _remove(self):
        self._grid.remove()

    @auto_refresh
    def set_xspacing(self, xspacing):
        '''
        Set the grid line spacing in the longitudinal direction

        Parameters
        ----------
        xspacing : { float, str }
            The spacing in the longitudinal direction. To set the spacing
            to be the same as the ticks, set this to 'tick'
        '''
        if xspacing == 'tick':
            self.x_auto_spacing = True
        elif np.isreal(xspacing):
            self.x_auto_spacing = False
            if self._wcs.xaxis_coord_type in ['longitude', 'latitude']:
                self.x_grid_spacing = au.Angle(
                    degrees=xspacing,
                    latitude=self._wcs.xaxis_coord_type == 'latitude')
            else:
                self.x_grid_spacing = xspacing
        else:
            raise ValueError("Grid spacing should be a scalar or 'tick'")

        self._update()

    @auto_refresh
    def set_yspacing(self, yspacing):
        '''
        Set the grid line spacing in the latitudinal direction

        Parameters
        ----------
        yspacing : { float, str }
            The spacing in the latitudinal direction. To set the spacing
            to be the same as the ticks, set this to 'tick'
        '''

        if yspacing == 'tick':
            self.y_auto_spacing = True
        elif np.isreal(yspacing):
            self.y_auto_spacing = False
            if self._wcs.yaxis_coord_type in ['longitude', 'latitude']:
                self.y_grid_spacing = au.Angle(
                    degrees=yspacing,
                    latitude=self._wcs.yaxis_coord_type == 'latitude')
            else:
                self.y_grid_spacing = yspacing
        else:
            raise ValueError("Grid spacing should be a scalar or 'tick'")

        self._update()

    @auto_refresh
    def set_color(self, color):
        '''
        Set the color of the grid lines

        Parameters
        ----------
        color : str
            The color of the grid lines
        '''
        if self._grid:
            self._grid.set_edgecolor(color)
        else:
            self.default_color = color

    @auto_refresh
    def set_alpha(self, alpha):
        '''
        Set the alpha (transparency) of the grid lines

        Parameters
        ----------
        alpha : float
            The alpha value of the grid. This should be a floating
            point value between 0 and 1, where 0 is completely
            transparent, and 1 is completely opaque.
        '''
        if self._grid:
            self._grid.set_alpha(alpha)
        else:
            self.default_alpha = alpha

    @auto_refresh
    def set_linewidth(self, linewidth):
        self._grid.set_linewidth(linewidth)

    @auto_refresh
    def set_linestyle(self, linestyle):
        self._grid.set_linestyle(linestyle)

    @auto_refresh
    def show(self):
        if self._grid:
            self._grid.set_visible(True)
        else:
            self._active = True
            self._update()
            self.set_color(self.default_color)
            self.set_alpha(self.default_alpha)

    @auto_refresh
    def hide(self):
        self._grid.set_visible(False)

    @auto_refresh
    def _update(self, *args):
        self._update_norefresh(*args)

    def _update_norefresh(self, *args):

        if not self._active:
            return self.ax

        if len(args) == 1:
            if id(self.ax) != id(args[0]):
                raise Exception("ax ids should match")

        lines = []

        # Set x grid spacing
        if self.x_auto_spacing:
            if self.ax.xaxis.apl_auto_tick_spacing:
                xspacing = default_spacing(self.ax, 'x',
                                           self.ax.xaxis.apl_label_form)
            else:
                xspacing = self.ax.xaxis.apl_tick_spacing
        else:
            xspacing = self.x_grid_spacing

        if xspacing is None:
            warnings.warn("Could not determine x tick spacing - grid cannot be drawn")
            return

        if self._wcs.xaxis_coord_type in ['longitude', 'latitude']:
            xspacing = xspacing.todegrees()

        # Set y grid spacing
        if self.y_auto_spacing:
            if self.ax.yaxis.apl_auto_tick_spacing:
                yspacing = default_spacing(self.ax, 'y',
                                           self.ax.yaxis.apl_label_form)
            else:
                yspacing = self.ax.yaxis.apl_tick_spacing
        else:
            yspacing = self.y_grid_spacing

        if yspacing is None:
            warnings.warn("Could not determine y tick spacing - grid cannot be drawn")
            return

        if self._wcs.yaxis_coord_type in ['longitude', 'latitude']:
            yspacing = yspacing.todegrees()

        # Find x lines that intersect with axes
        grid_x_i, grid_y_i = find_intersections(self.ax, 'x', xspacing)

        # Ensure that longitudes are between 0 and 360, and latitudes between
        # -90 and 90
        if self._wcs.xaxis_coord_type == 'longitude':
            grid_x_i = np.mod(grid_x_i, 360.)
        elif self._wcs.xaxis_coord_type == 'latitude':
            grid_x_i = np.mod(grid_x_i + 90., 180.) - 90.

        if self._wcs.yaxis_coord_type == 'longitude':
            grid_y_i = np.mod(grid_y_i, 360.)
        elif self._wcs.yaxis_coord_type == 'latitude':
            grid_y_i = np.mod(grid_y_i + 90., 180.) - 90.

        # If we are dealing with longitude/latitude then can search all
        # neighboring grid lines to see if there are any closed longitude
        # lines
        if self._wcs.xaxis_coord_type == 'latitude' and self._wcs.yaxis_coord_type == 'longitude' and len(grid_x_i) > 0:

            gx = grid_x_i.min()

            while True:
                gx -= xspacing
                xpix, ypix = wcs_util.world2pix(self._wcs, gx, 0.)
                if in_plot(self.ax, xpix, ypix) and gx >= -90.:
                    grid_x_i = np.hstack([grid_x_i, gx, gx])
                    grid_y_i = np.hstack([grid_y_i, 0., 360.])
                else:
                    break

            gx = grid_x_i.max()

            while True:
                gx += xspacing
                xpix, ypix = wcs_util.world2pix(self._wcs, gx, 0.)
                if in_plot(self.ax, xpix, ypix) and gx <= +90.:
                    grid_x_i = np.hstack([grid_x_i, gx, gx])
                    grid_y_i = np.hstack([grid_y_i, 0., 360.])
                else:
                    break

        # Plot those lines
        for gx in np.unique(grid_x_i):
            for line in plot_grid_x(self.ax, grid_x_i, grid_y_i, gx):
                lines.append(line)

        # Find y lines that intersect with axes
        grid_x_i, grid_y_i = find_intersections(self.ax, 'y', yspacing)

        if self._wcs.xaxis_coord_type == 'longitude':
            grid_x_i = np.mod(grid_x_i, 360.)
        elif self._wcs.xaxis_coord_type == 'latitude':
            grid_x_i = np.mod(grid_x_i + 90., 180.) - 90.

        if self._wcs.yaxis_coord_type == 'longitude':
            grid_y_i = np.mod(grid_y_i, 360.)
        elif self._wcs.yaxis_coord_type == 'latitude':
            grid_y_i = np.mod(grid_y_i + 90., 180.) - 90.

        # If we are dealing with longitude/latitude then can search all
        # neighboring grid lines to see if there are any closed longitude
        # lines
        if (self._wcs.xaxis_coord_type == 'longitude' and
           self._wcs.yaxis_coord_type == 'latitude' and len(grid_y_i) > 0):

            gy = grid_y_i.min()

            while True:
                gy -= yspacing
                xpix, ypix = wcs_util.world2pix(self._wcs, 0., gy)
                if in_plot(self.ax, xpix, ypix) and gy >= -90.:
                    grid_x_i = np.hstack([grid_x_i, 0., 360.])
                    grid_y_i = np.hstack([grid_y_i, gy, gy])
                else:
                    break

            gy = grid_y_i.max()

            while True:
                gy += yspacing
                xpix, ypix = wcs_util.world2pix(self._wcs, 0., gy)
                if in_plot(self.ax, xpix, ypix) and gy <= +90.:
                    grid_x_i = np.hstack([grid_x_i, 0., 360.])
                    grid_y_i = np.hstack([grid_y_i, gy, gy])
                else:
                    break

        # Plot those lines
        for gy in np.unique(grid_y_i):
            for line in plot_grid_y(self.ax, grid_x_i, grid_y_i, gy):
                lines.append(line)

        if self._grid:
            self._grid.set_verts(lines)
        else:
            self._grid = LineCollection(lines, transOffset=self.ax.transData)
            self.ax.add_collection(self._grid, False)

        return self.ax


def plot_grid_y(ax, grid_x, grid_y, gy, alpha=0.5):
    '''Plot a single grid line in the y direction'''

    wcs = ax._wcs
    lines_out = []

    # Find intersections that correspond to latitude lat0
    index = np.where(grid_y == gy)

    # Produce sorted array of the longitudes of all intersections
    grid_x_sorted = np.sort(grid_x[index])

    # If coordinate type is a latitude or longitude, also need to check if
    # end-points fall inside the plot
    if wcs.xaxis_coord_type == 'latitude':
        if not np.any(grid_x_sorted == -90):
            xpix, ypix = wcs_util.world2pix(wcs, max(grid_x_sorted[0] - 1., -90.), gy)
            if in_plot(ax, xpix, ypix):
                grid_x_sorted = np.hstack([-90., grid_x_sorted])
        if not np.any(grid_x_sorted == +90):
            xpix, ypix = wcs_util.world2pix(wcs, min(grid_x_sorted[-1] + 1., +90.), gy)
            if in_plot(ax, xpix, ypix):
                grid_x_sorted = np.hstack([grid_x_sorted, +90.])
    elif wcs.xaxis_coord_type == 'longitude':
        if not np.any(grid_x_sorted == 0.):
            xpix, ypix = wcs_util.world2pix(wcs, max(grid_x_sorted[0] - 1., 0.), gy)
            if in_plot(ax, xpix, ypix):
                grid_x_sorted = np.hstack([0., grid_x_sorted])
        if not np.any(grid_x_sorted == 360.):
            xpix, ypix = wcs_util.world2pix(wcs, min(grid_x_sorted[-1] + 1., 360.), gy)
            if in_plot(ax, xpix, ypix):
                grid_x_sorted = np.hstack([grid_x_sorted, 360.])

    # Check if the first mid-point with coordinates is inside the viewport
    xpix, ypix = wcs_util.world2pix(wcs, (grid_x_sorted[0] + grid_x_sorted[1]) / 2., gy)

    if not in_plot(ax, xpix, ypix):
        grid_x_sorted = np.roll(grid_x_sorted, 1)

    # Check that number of grid points is even
    if len(grid_x_sorted) % 2 == 1:
        warnings.warn("Unexpected number of grid points - x grid lines cannot be drawn")
        return []

    # Cycle through intersections
    for i in range(0, len(grid_x_sorted), 2):

        grid_x_min = grid_x_sorted[i]
        grid_x_max = grid_x_sorted[i + 1]

        x_world = math_util.complete_range(grid_x_min, grid_x_max, 100)
        y_world = np.repeat(gy, len(x_world))
        x_pix, y_pix = wcs_util.world2pix(wcs, x_world, y_world)
        lines_out.append(list(zip(x_pix, y_pix)))

    return lines_out


def plot_grid_x(ax, grid_x, grid_y, gx, alpha=0.5):
    '''Plot a single longitude line'''

    wcs = ax._wcs
    lines_out = []

    # Find intersections that correspond to longitude gx
    index = np.where(grid_x == gx)

    # Produce sorted array of the latitudes of all intersections
    grid_y_sorted = np.sort(grid_y[index])

    # If coordinate type is a latitude or longitude, also need to check if
    # end-points fall inside the plot
    if wcs.yaxis_coord_type == 'latitude':
        if not np.any(grid_y_sorted == -90):
            xpix, ypix = wcs_util.world2pix(wcs, gx, max(grid_y_sorted[0] - 1., -90.))
            if in_plot(ax, xpix, ypix):
                grid_y_sorted = np.hstack([-90., grid_y_sorted])
        if not np.any(grid_y_sorted == +90):
            xpix, ypix = wcs_util.world2pix(wcs, gx, min(grid_y_sorted[-1] + 1., +90.))
            if in_plot(ax, xpix, ypix):
                grid_y_sorted = np.hstack([grid_y_sorted, +90.])
    elif wcs.yaxis_coord_type == 'longitude':
        if not np.any(grid_y_sorted == 0.):
            xpix, ypix = wcs_util.world2pix(wcs, gx, max(grid_y_sorted[0] - 1., 0.))
            if in_plot(ax, xpix, ypix):
                grid_y_sorted = np.hstack([0., grid_y_sorted])
        if not np.any(grid_y_sorted == 360.):
            xpix, ypix = wcs_util.world2pix(wcs, gx, min(grid_y_sorted[-1] + 1., 360.))
            if in_plot(ax, xpix, ypix):
                grid_y_sorted = np.hstack([grid_y_sorted, 360.])

    # Check if the first mid-point with coordinates is inside the viewport
    xpix, ypix = wcs_util.world2pix(wcs, gx, (grid_y_sorted[0] + grid_y_sorted[1]) / 2.)

    if not in_plot(ax, xpix, ypix):
        grid_y_sorted = np.roll(grid_y_sorted, 1)

    # Check that number of grid points is even
    if len(grid_y_sorted) % 2 == 1:
        warnings.warn("Unexpected number of grid points - y grid lines cannot be drawn")
        return []

    # Cycle through intersections
    for i in range(0, len(grid_y_sorted), 2):

        grid_y_min = grid_y_sorted[i]
        grid_y_max = grid_y_sorted[i + 1]

        y_world = math_util.complete_range(grid_y_min, grid_y_max, 100)
        x_world = np.repeat(gx, len(y_world))
        x_pix, y_pix = wcs_util.world2pix(wcs, x_world, y_world)
        lines_out.append(list(zip(x_pix, y_pix)))

    return lines_out


def in_plot(ax, x_pix, y_pix):
    '''Check whether a given point is in a plot'''

    xmin, xmax = ax.xaxis.get_view_interval()
    ymin, ymax = ax.yaxis.get_view_interval()

    return (x_pix > xmin + 0.5 and x_pix < xmax + 0.5 and
            y_pix > ymin + 0.5 and y_pix < ymax + 0.5)


def find_intersections(ax, coord, spacing):
    '''
    Find intersections of a given coordinate with all axes

    Parameters
    ----------

    ax :
       The matplotlib axis instance for the figure.

    coord : { 'x', 'y' }
       The coordinate for which we are looking for ticks.

    spacing : float
       The spacing along the axis.

    '''

    wcs = ax._wcs

    xmin, xmax = ax.xaxis.get_view_interval()
    ymin, ymax = ax.yaxis.get_view_interval()

    options = dict(mode='xy', xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)

    # Initialize arrays
    x, y = [], []

    # Bottom X axis
    (labels_x, labels_y, world_x, world_y) = tick_positions(
        wcs, spacing, 'x', coord, farside=False, **options)

    x.extend(world_x)
    y.extend(world_y)

    # Top X axis
    (labels_x, labels_y, world_x, world_y) = tick_positions(
        wcs, spacing, 'x', coord, farside=True, **options)

    x.extend(world_x)
    y.extend(world_y)

    # Left Y axis
    (labels_x, labels_y, world_x, world_y) = tick_positions(
        wcs, spacing, 'y', coord, farside=False, **options)

    x.extend(world_x)
    y.extend(world_y)

    # Right Y axis
    (labels_x, labels_y, world_x, world_y) = tick_positions(
        wcs, spacing, 'y', coord, farside=True, **options)

    x.extend(world_x)
    y.extend(world_y)

    return np.array(x), np.array(y)



import os

import matplotlib
matplotlib.use('Agg')

import numpy as np
from astropy.tests.helper import pytest
from astropy.io import fits
from astropy.wcs import WCS as AstropyWCS

from .helpers import generate_file, generate_hdu, generate_wcs
from .. import FITSFigure


def test_convolve_default():
    data = np.random.random((16, 16))
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale(smooth=3)
    f.close()


def test_convolve_gauss():
    data = np.random.random((16, 16))
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale(kernel='gauss', smooth=3)
    f.close()


def test_convolve_box():
    data = np.random.random((16, 16))
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale(kernel='box', smooth=3)
    f.close()


def test_convolve_custom():
    data = np.random.random((16, 16))
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale(kernel=np.ones((3,3)))
    f.close()


def test_convolve_default():
    # Regression test for aplpy/aplpy#165
    data = np.ones((16, 16), dtype=int)
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale(smooth=3)
    f.close()



import os
import warnings

import matplotlib
matplotlib.use('Agg')

import numpy as np
from astropy.io import fits

from .. import FITSFigure
from ..rgb import make_rgb_image

from .test_images import BaseImageTests

HEADER = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/2d_fits', '1904-66_TAN.hdr')


class TestRGB(BaseImageTests):

    def test_rgb(self, generate, tmpdir):

        # Regression test to check that RGB recenter works properly

        r_file = tmpdir.join('r.fits').strpath
        g_file = tmpdir.join('g.fits').strpath
        b_file = tmpdir.join('b.fits').strpath
        rgb_file = tmpdir.join('rgb.png').strpath

        np.random.seed(12345)

        header = fits.Header.fromtextfile(HEADER)

        r = fits.PrimaryHDU(np.random.random((12,12)), header)
        r.writeto(r_file)

        g = fits.PrimaryHDU(np.random.random((12,12)), header)
        g.writeto(g_file)

        b = fits.PrimaryHDU(np.random.random((12,12)), header)
        b.writeto(b_file)

        make_rgb_image([r_file, g_file, b_file], rgb_file, embed_avm_tags=False)

        f = FITSFigure(r_file, figsize=(3,3))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            f.show_rgb(rgb_file)

        f.tick_labels.set_xformat('dd.d')
        f.tick_labels.set_yformat('dd.d')

        f.recenter(359.3, -72.1, radius=0.05)

        self.generate_or_test(generate, f, 'test_rgb.png', tolerance=2)
        f.close()



from __future__ import absolute_import, print_function, division

import numpy as np


def isnumeric(value):
    return type(value) in [float, int, np.int8, np.int16, np.int32, \
        np.float32, np.float64]


def smart_range(array):

    array.sort()

    minval = 360.
    i1 = 0
    i2 = 0

    for i in range(0, np.size(array) - 1):
        if 360. - abs(array[i + 1] - array[i]) < minval:
            minval = 360. - abs(array[i + 1] - array[i])
            i1 = i + 1
            i2 = i

    if(max(array) - min(array) < minval):
        i1 = 0
        i2 = np.size(array) - 1

    x_min = array[i1]
    x_max = array[i2]

    if(x_min > x_max):
        x_min = x_min - 360.

    return x_min, x_max


def complete_range(xmin, xmax, spacing):
    if(xmax - xmin < 1):
        spacing = 10
    xstep = (xmax - xmin) / float(spacing)
    r = np.arange(xmin, xmax, xstep)
    if(np.any(r >= xmax)):
        return r
    else:
        return np.hstack([r, xmax])


def closest(array, a):
    ipos = np.argmin(np.abs(a - array))
    return array[ipos]


def divisors(n, dup=False):
    divisors = []
    i = 0
    if n == 1:
        return {1: 1}
    while 1:
        i += 1
        if dup:
            break
        if i == n + 1:
            break
        if n % i == 0:
            divisors[i:i + 1] = [i]
    return np.array(divisors)



import numpy as np

from ..core import FITSFigure

from .test_images import BaseImageTests

x = np.linspace(-1., 1., 10)
y = np.linspace(-1., 1., 10)
X, Y = np.meshgrid(x, y)

np.random.seed(12345)
IMAGE = np.random.random((10,10))
PDATA = np.arange(100).reshape((10,10)) / 50. + 0.5
ADATA = np.degrees(np.arctan2(Y, X))

class TestVectors(BaseImageTests):

    def test_default(self, generate):
        f = FITSFigure(IMAGE, figsize=(4,4))
        f.show_grayscale()
        f.show_vectors(PDATA, ADATA, color='orange')
        self.generate_or_test(generate, f, 'vectors_default.png', tolerance=2.5)
        f.close()

    def test_step_scale(self, generate):
        f = FITSFigure(IMAGE, figsize=(4,4))
        f.show_grayscale()
        f.show_vectors(PDATA, ADATA, step=2, scale=0.8, color='orange')
        self.generate_or_test(generate, f, 'vectors_step_scale.png', tolerance=2.5)
        f.close()




import os

import numpy as np
from astropy.io import fits
from astropy.tests.helper import pytest

from ..wcs_util import WCS, celestial_pixel_scale, non_celestial_pixel_scales

from .helpers import generate_wcs

HEADER_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
HEADER_2D = fits.Header.fromtextfile(os.path.join(HEADER_DIR, '2d_fits', '1904-66_TAN.hdr'))
HEADER_3D = fits.Header.fromtextfile(os.path.join(HEADER_DIR, '3d_fits', 'cube.hdr'))


def test_pixel_scale_matrix_1():
    wcs = WCS(HEADER_2D)
    np.testing.assert_allclose(wcs.pixel_scale_matrix, [[-0.06666667, 0.], [0., 0.06666667]])


def test_celestial_scale_1():
    wcs = WCS(HEADER_2D)
    for crota2 in range(0, 360, 20):
        wcs.wcs.crota = [0.,crota2]
        np.testing.assert_allclose(celestial_pixel_scale(wcs), 0.06666667)


def test_non_celestial_scales_1():
    wcs = WCS(HEADER_2D)
    with pytest.raises(ValueError) as exc:
        non_celestial_pixel_scales(wcs)
    assert exc.value.args[0] == "WCS is celestial, use celestial_pixel_scale instead"


def test_pixel_scale_matrix_2():
    wcs = WCS(HEADER_3D, dimensions=[0,2], slices=[0])
    np.testing.assert_allclose(wcs.pixel_scale_matrix, [[-6.38888900e-03, 0.], [0., 6.64236100e+01]])


def test_celestial_scale_2():
    wcs = WCS(HEADER_3D, dimensions=[0,2], slices=[0])
    with pytest.raises(ValueError) as exc:
        celestial_pixel_scale(wcs)
    assert exc.value.args[0] == "WCS is not celestial, cannot determine celestial pixel scale"


def test_non_celestial_scales_2():
    wcs = WCS(HEADER_3D, dimensions=[0,2], slices=[0])
    np.testing.assert_allclose(non_celestial_pixel_scales(wcs), [6.38888900e-03, 6.64236100e+01])



from __future__ import absolute_import, print_function, division

from matplotlib.font_manager import FontProperties

from . import wcs_util
from .decorators import auto_refresh, fixdocstring


class AxisLabels(object):

    def __init__(self, parent):

        # Store references to axes
        self._ax1 = parent._ax1
        self._ax2 = parent._ax2
        self._wcs = parent._wcs
        self._figure = parent._figure

        # Save plotting parameters (required for @auto_refresh)
        self._parameters = parent._parameters

        # Set font
        self._label_fontproperties = FontProperties()

        self._ax2.yaxis.set_label_position('right')
        self._ax2.xaxis.set_label_position('top')

        system, equinox, units = wcs_util.system(self._wcs)

        if system['name'] == 'equatorial':

            if equinox == 'b1950':
                xtext = 'RA (B1950)'
                ytext = 'Dec (B1950)'
            else:
                xtext = 'RA (J2000)'
                ytext = 'Dec (J2000)'

        elif system['name'] == 'galactic':

            xtext = 'Galactic Longitude'
            ytext = 'Galactic Latitude'

        elif system['name'] == 'ecliptic':

            xtext = 'Ecliptic Longitude'
            ytext = 'Ecliptic Latitude'

        elif system['name'] == 'unknown':

            xunit = " (%s)" % self._wcs.cunit_x if self._wcs.cunit_x not in ["", None] else ""
            yunit = " (%s)" % self._wcs.cunit_y if self._wcs.cunit_y not in ["", None] else ""

            if len(self._wcs.cname_x) > 0:
                xtext = self._wcs.cname_x + xunit
            else:
                if len(self._wcs.ctype_x) == 8 and self._wcs.ctype_x[4] == '-':
                    xtext = self._wcs.ctype_x[:4].replace('-', '') + xunit
                else:
                    xtext = self._wcs.ctype_x + xunit

            if len(self._wcs.cname_y) > 0:
                ytext = self._wcs.cname_y + yunit
            else:
                if len(self._wcs.ctype_y) == 8 and self._wcs.ctype_y[4] == '-':
                    ytext = self._wcs.ctype_y[:4].replace('-', '') + yunit
                else:
                    ytext = self._wcs.ctype_y + yunit

        if system['inverted']:
            xtext, ytext = ytext, xtext

        self.set_xtext(xtext)
        self.set_ytext(ytext)

        self.set_xposition('bottom')
        self.set_yposition('left')

    @auto_refresh
    def set_xtext(self, label):
        """
        Set the x-axis label text.
        """
        self._xlabel1 = self._ax1.set_xlabel(label)
        self._xlabel2 = self._ax2.set_xlabel(label)

    @auto_refresh
    def set_ytext(self, label):
        """
        Set the y-axis label text.
        """
        self._ylabel1 = self._ax1.set_ylabel(label)
        self._ylabel2 = self._ax2.set_ylabel(label)

    @auto_refresh
    def set_xpad(self, pad):
        """
        Set the x-axis label displacement, in points.
        """
        self._xlabel1 = self._ax1.set_xlabel(self._xlabel1.get_text(), labelpad=pad)
        self._xlabel2 = self._ax2.set_xlabel(self._xlabel2.get_text(), labelpad=pad)

    @auto_refresh
    def set_ypad(self, pad):
        """
        Set the y-axis label displacement, in points.
        """
        self._ylabel1 = self._ax1.set_ylabel(self._ylabel1.get_text(), labelpad=pad)
        self._ylabel2 = self._ax2.set_ylabel(self._ylabel2.get_text(), labelpad=pad)

    @auto_refresh
    @fixdocstring
    def set_font(self, family=None, style=None, variant=None, stretch=None, weight=None, size=None, fontproperties=None):
        """
        Set the font of the axis labels.

        Parameters
        ----------

        common: family, style, variant, stretch, weight, size, fontproperties

        Notes
        -----

        Default values are set by matplotlib or previously set values if
        set_font has already been called. Global default values can be set by
        editing the matplotlibrc file.
        """

        if family:
            self._label_fontproperties.set_family(family)

        if style:
            self._label_fontproperties.set_style(style)

        if variant:
            self._label_fontproperties.set_variant(variant)

        if stretch:
            self._label_fontproperties.set_stretch(stretch)

        if weight:
            self._label_fontproperties.set_weight(weight)

        if size:
            self._label_fontproperties.set_size(size)

        if fontproperties:
            self._label_fontproperties = fontproperties

        self._xlabel1.set_fontproperties(self._label_fontproperties)
        self._xlabel2.set_fontproperties(self._label_fontproperties)
        self._ylabel1.set_fontproperties(self._label_fontproperties)
        self._ylabel2.set_fontproperties(self._label_fontproperties)

    @auto_refresh
    def show(self):
        """
        Show the x- and y-axis labels.
        """
        self.show_x()
        self.show_y()

    @auto_refresh
    def hide(self):
        """
        Hide the x- and y-axis labels.
        """
        self.hide_x()
        self.hide_y()

    @auto_refresh
    def show_x(self):
        """
        Show the x-axis label.
        """
        if self._xposition == 'bottom':
            self._xlabel1.set_visible(True)
        else:
            self._xlabel2.set_visible(True)

    @auto_refresh
    def hide_x(self):
        """
        Hide the x-axis label.
        """
        if self._xposition == 'bottom':
            self._xlabel1.set_visible(False)
        else:
            self._xlabel2.set_visible(False)

    @auto_refresh
    def show_y(self):
        """
        Show the y-axis label.
        """
        if self._yposition == 'left':
            self._ylabel1.set_visible(True)
        else:
            self._ylabel2.set_visible(True)

    @auto_refresh
    def hide_y(self):
        """
        Hide the y-axis label.
        """
        if self._yposition == 'left':
            self._ylabel1.set_visible(False)
        else:
            self._ylabel2.set_visible(False)

    @auto_refresh
    def set_xposition(self, position):
        "Set the position of the x-axis label ('top' or 'bottom')"
        if position == 'bottom':
            self._xlabel1.set_visible(True)
            self._xlabel2.set_visible(False)
        elif position == 'top':
            self._xlabel1.set_visible(False)
            self._xlabel2.set_visible(True)
        else:
            raise ValueError("position should be one of 'top' or 'bottom'")
        self._xposition = position

    @auto_refresh
    def set_yposition(self, position):
        "Set the position of the y-axis label ('left' or 'right')"
        if position == 'left':
            self._ylabel1.set_visible(True)
            self._ylabel2.set_visible(False)
        elif position == 'right':
            self._ylabel1.set_visible(False)
            self._ylabel2.set_visible(True)
        else:
            raise ValueError("position should be one of 'left' or 'right'")
        self._yposition = position



import os

import matplotlib
matplotlib.use('Agg')

import numpy as np

from astropy.table import Table
from astropy.tests.helper import pytest
from .helpers import generate_wcs
from .. import wcs_util


HEADER = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/2d_fits', '1904-66_TAN.hdr')

tab = Table({'RA':[347.,349.], 'DEC':[-68.,-68]})

GOOD_INPUT = [ [1.,2.],
               [[1.,2.],[3,4]],
               [np.arange(2), np.arange(2)],
               [tab['RA'], tab['DEC']]
             ]

BAD_INPUT = [ [1,['s','w']],
              [np.arange(2), np.sum],
              [tab['RA'], 'ewr']
            ]

@pytest.mark.parametrize(('inputval'), GOOD_INPUT)
def test_pixworld_input_and_consistency(inputval):
    wcs = generate_wcs(HEADER)
    ra, dec = wcs_util.pix2world(wcs, inputval[0], inputval[1])
    x, y = wcs_util.world2pix(wcs, ra, dec)
    # For some inputs (e.g. list) a-b is not defined
    # so change them all to np.ndarrays here to make sure
    assert np.all(np.abs(np.array(x) - np.array(inputval[0])) < 1e-5)
    assert np.all(np.abs(np.array(y) - np.array(inputval[1])) < 1e-5)

def test_returntypes():
    wcs = generate_wcs(HEADER)
    ra, dec = wcs_util.pix2world(wcs, 1.,2.)
    assert np.isscalar(ra) and np.isscalar(dec)
    ra, dec = wcs_util.pix2world(wcs, [1.],[2.])
    assert (type(ra) == list) and (type(dec) == list)
    # Astropy.table.Column objects get donwconverted np.ndarray
    ra, dec = wcs_util.pix2world(wcs, np.arange(2), tab['DEC'])
    assert isinstance(ra, np.ndarray) and isinstance(dec, np.ndarray)


@pytest.mark.parametrize(('inputval'), BAD_INPUT)
def test_pix2world_fail(inputval):
    wcs = generate_wcs(HEADER)
    with pytest.raises(Exception) as exc:
        wcs_util.pix2world(wcs, inputval[0], inputval[1])
    assert exc.value.args[0] == "pix2world should be provided either with two scalars, two lists, or two numpy arrays"

@pytest.mark.parametrize(('inputval'), BAD_INPUT)
def test_world2pix_fail(inputval):
    wcs = generate_wcs(HEADER)
    with pytest.raises(Exception) as exc:
        wcs_util.world2pix(wcs, inputval[0], inputval[1])
    assert exc.value.args[0] == "world2pix should be provided either with two scalars, two lists, or two numpy arrays"

